"""
Copyright 2017 Deepgram

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import math
import bisect

import numpy

###############################################################################
def argsort(data, batch_size, neighborhood=None, growth_factor=None,
	uniform=False):
	""" Returns indices which sort the dataset into batches of neighboring
		entries.

		# Arguments

		data: list or numpy array. The data to use in producing the sort
			indices.
		batch_size: int. The size of each batch.
		neighborhood: int or float (default: None). The target size to draw
			neighboring samples from. If None, it defaults to 10% of the range
			of `data`.
		growth_factor: float (default: None). If there are not enough data
			points in the neighborhood to construct an entire batch, then the
			neighborhood must expand to include other, more distant points.
			This parameter controls how much the current neighborhood should
			exand when necessary. It is a fraction of the current neighborhood
			size. If this is None, it defaults to 0.4.
		uniform: bool (default: False). If True, candidate neighborhoods are
			selected uniformly at random from all data points not previously
			associated with a neighborhood. If False, neighborhoods are
			systematically "grown" from one end of the data towards the other.

		# Return value

		A numpy array of int32's that contains the indices that would sort the
		array into neighborhoods.
	"""

	num_entries = len(data)
	num_batches = math.ceil(num_entries / batch_size)
	num_perfect_batches = int(num_entries / batch_size)

	# Sort the data.
	sort_permutation = numpy.argsort(data)
	data = data[sort_permutation]

	# Calculate the neighborhood size if it wasn't given to us.
	if not neighborhood:
		data_range = data[-1] - data[0]
		neighborhood = data_range / 10

	if not uniform:
		neighborhood *= 2

	if not growth_factor:
		growth_factor = 0.4

	# This mask tracks which entries we've already assigned to neighborhoods.
	mask = numpy.ones(num_entries, dtype=bool)

	# We will put our results here.
	result = numpy.empty(num_entries, dtype=numpy.int32)

	if not uniform:
		batch_ordering = numpy.random.permutation(num_perfect_batches)

	# Create each batch
	for num_batch in range(num_batches):

		offset = batch_size * num_batch
		needed = min(num_entries - offset, batch_size)

		# Choose our target "value" for the neighborhood.
		if uniform:
			seed_index = numpy.random.choice(num_entries - offset)
		else:
			if numpy.random.random_sample() >= 0.5:
				seed_index = num_entries - offset - 1
			else:
				seed_index = 0

		seed_value = data[mask][seed_index]

		# Find a neighborhood that has enough data.
		current_neighborhood = neighborhood / 2
		while True:
			lower_bound = bisect.bisect_left(
				data[mask],
				seed_value - current_neighborhood
			)
			upper_bound = bisect.bisect_right(
				data[mask],
				seed_value + current_neighborhood
			)

			candidates = upper_bound - lower_bound
			if candidates < needed:
				current_neighborhood *= (1 + growth_factor)
			else:
				break

		# Select which points to assign to this neighborhood.
		selected = numpy.random.choice(
			candidates, size=needed, replace=False
		) + lower_bound

		# Keep the seed.
		if not uniform:
			if seed_index not in selected:
				selected[0] = seed_index

		# Update the results.
		if not uniform and num_batch < num_perfect_batches:
			# For non-uniform sampling, we still want the end results in a
			# pseudo-random order.
			offset = batch_ordering[num_batch]*batch_size

		result[offset:offset+needed] = sort_permutation[mask][selected]

		# Update the mask.
		mask[numpy.nonzero(mask)[0][selected]] = False

	return result

### EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF



"""
Copyright 2017 Deepgram

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import logging

from . import Layer, ParsingError

logger = logging.getLogger(__name__)

###############################################################################
class Dropout(Layer):	# pylint: disable=too-few-public-methods
	""" A dropout layer.
	"""

	DEFAULT_INDEPENDENT = True

	###########################################################################
	def __init__(self, *args, **kwargs):
		""" Creates a new dropout layer.
		"""
		super().__init__(*args, **kwargs)
		self.dropout = None
		self.independent = None

	###########################################################################
	def _parse(self, engine):
		""" Parse the layer.
		"""
		if isinstance(self.args, dict):
			if 'fraction' in self.args:
				self.dropout = engine.evaluate(
					self.args['fraction'], recursive=True)
				if not isinstance(self.dropout, (int, float)):
					raise ParsingError('"dropout" must be a float.')
			else:
				raise ParsingError('Missing required key "fraction" for '
					'Dropout layer.')

			if 'independent' in self.args:
				self.independent = engine.evaluate(self.args['independent'],
					recursive=True)
				if not isinstance(self.independent, bool):
					raise ParsingError('"independent" must be boolean.')
			else:
				self.independent = self.DEFAULT_INDEPENDENT
		elif isinstance(self.args, (int, float)):
			self.dropout = self.args
			self.independent = self.DEFAULT_INDEPENDENT
		else:
			raise ParsingError('Dropout layer requires a single '
				'floating-point argument argument. Instead we received: {}'
				.format(self.args))

		self.dropout = float(self.dropout)
		if self.dropout < 0 or self.dropout >= 1:
			raise ParsingError('Dropout fraction must be >= 0 and < 1.')

	###########################################################################
	def _build(self, model):
		""" Instantiates the layer with the given backend.
		"""
		backend = model.get_backend()
		if backend.get_name() == 'keras':

			if not self.independent:
				logger.warning('Keras backend only supports independent '
					'dropout. Pretending that "independent" is True.')

			import keras.layers as L			# pylint: disable=import-error
			yield L.Dropout(
				self.dropout,
				name=self.name
			)

		elif backend.get_name() == 'pytorch':

			import torch.nn as nn				# pylint: disable=import-error
			from kur.backend.pytorch.modules import swap_channels

			def connect(inputs):
				""" Connects the layer
				"""
				assert len(inputs) == 1

				ndim = len(inputs[0]['shape'])
				func = nn.Dropout if self.independent else {
					2 : nn.Dropout2d,
					3 : nn.Dropout3d
				}.get(ndim-1, nn.Dropout)

				output = inputs[0]['layer']

				if func is not nn.Dropout:
					output = model.data.add_operation(
						swap_channels
					)(output)

				output = model.data.add_layer(
					self.name,
					func(self.dropout)
				)(output)

				if func is not nn.Dropout:
					output = model.data.add_operation(
						swap_channels
					)(output)

				return {
					'shape' : self.shape([inputs[0]['shape']]),
					'layer' : output
				}

			yield connect

		else:
			raise ValueError(
				'Unknown or unsupported backend: {}'.format(backend))

	###########################################################################
	def shape(self, input_shapes):
		""" Returns the output shape of this layer for a given input shape.
		"""
		if len(input_shapes) > 1:
			raise ValueError('Dropout only take a single input.')
		input_shape = input_shapes[0]
		return input_shape

### EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF



"""
Copyright 2016 Deepgram

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from . import Layer, ParsingError

###############################################################################
class Pooling(Layer):				# pylint: disable=too-few-public-methods
	""" A pooling layer

		# Properties

		size: int or list of ints (required). The size of each kernel. If an
			integer is used, it is interpretted as a one-dimensional
			convolution, the same as if it were put into a length-1 list.
		strides: int or list of ints (optional; default: 1 in each dimension).
			The stride (subsampling rate) between convolutions. If a list, it
			must be the same length as `size` and specify the stride in each
			respective dimension. If a single number, it is used as the stride
			in each dimension.
		pool: one of (max, average). The pool function to apply.

		# Example

		```
		pool:
		  size: [2, 2]
		  strides: 1
		  type: max
		```
	"""

	POOL_TYPES = ('max', 'average')

	###########################################################################
	@classmethod
	def get_container_name(cls):
		""" Returns the name of the container class.
		"""
		return 'pool'

	###########################################################################
	def __init__(self, *args, **kwargs):
		""" Creates a new pooling layer.
		"""
		super().__init__(*args, **kwargs)
		self.size = None
		self.strides = None
		self.pooltype = None

	###########################################################################
	def _parse(self, engine):
		""" Parses out the pooling layer.
		"""
		# Parse self
		if isinstance(self.args, dict):
			if 'size' not in self.args:
				raise ParsingError('Missing key "size" in pooling container.')
			self.size = engine.evaluate(self.args['size'], recursive=True)

			if 'strides' in self.args:
				self.strides = engine.evaluate(self.args['strides'],
					recursive=True)

			if 'type' in self.args:
				self.pooltype = engine.evaluate(self.args['type']).lower()
				if self.pooltype not in Pooling.POOL_TYPES:
					raise ParsingError('Unknown pool type "{}". Pool type can '
						'be one of: {}'.format(
							self.pooltype, ', '.join(Pooling.POOL_TYPES)
						))
		else:
			self.size = engine.evaluate(self.args, recursive=True)

		if self.pooltype is None:
			self.pooltype = Pooling.POOL_TYPES[0]

		if not isinstance(self.size, (list, tuple)):
			self.size = [self.size]
		if not 1 <= len(self.size) <= 3:
			raise ParsingError('Only pooling layers with dimensions 1, 2, '
				'or 3 are supported.')
		for i in range(len(self.size)):
			try:
				self.size[i] = int(self.size[i])
			except ValueError:
				raise ParsingError('All "size" entries must evaluate to '
					'integers. We received this instead: {}'
					.format(self.size[i]))

		if self.strides is not None:
			if not isinstance(self.strides, (list, tuple)):
				try:
					self.strides = int(self.strides)
				except ValueError:
					raise ParsingError('"strides" must evaluate to an '
						'integer or a list of integers.')
				self.strides = [self.strides] * len(self.size)
			else:
				if len(self.strides) != len(self.size):
					raise ParsingError('If "strides" is a list, it must '
						'be the same length as "size".')
				for i in range(len(self.strides)):
					try:
						self.strides[i] = int(self.strides[i])
					except ValueError:
						raise ParsingError('Each element of "strides" '
							'must evaluate to an integer.')
		else:
			self.strides = [1] * len(self.size)

	###########################################################################
	def _build(self, model):
		""" Instantiates the layer with the given backend.
		"""
		backend = model.get_backend()
		if backend.get_name() == 'keras':

			if backend.keras_version() == 1:
				import keras.layers as L		# pylint: disable=import-error
				kwargs = {
					'pool_size' : self.size,
					'strides' : self.strides,
					'border_mode' : 'valid',
					'name' : self.name
				}
				if len(self.size) == 1:
					kwargs['pool_length'] = kwargs.pop('pool_size')
					kwargs['stride'] = kwargs.pop('strides')

			else:
				import keras.layers.pooling as L # pylint: disable=import-error
				kwargs = {
					'pool_size' : self.size,
					'strides' : self.strides,
					'padding' : 'valid',
					'name' : self.name
				}
				if len(self.size) == 1:
					kwargs['pool_size'] = kwargs.pop('pool_size')[0]
				else:
					kwargs['data_format'] = 'channels_last'

			if self.pooltype == 'max':
				func = {
					1 : L.MaxPooling1D,
					2 : L.MaxPooling2D,
					3 : L.MaxPooling3D
				}.get(len(self.size))
			elif self.pooltype == 'average':
				func = {
					1 : L.AveragePooling1D,
					2 : L.AveragePooling2D,
					3 : L.AveragePooling3D
				}.get(len(self.size))
			else:
				raise ValueError('Unhandled pool type "{}". This is a bug.',
					self.pooltype)

			if func is None:
				raise ValueError('Invalid pool function for pool type "{}" '
					'the supplied pool parameters. This is a bug.'
					.format(self.pooltype))

			yield func(**kwargs)

		elif backend.get_name() == 'pytorch':

			import torch.nn as nn				# pylint: disable=import-error

			from kur.backend.pytorch.modules import swap_channels

			if self.pooltype == 'max':
				func = {
					1 : nn.MaxPool1d,
					2 : nn.MaxPool2d,
					3 : nn.MaxPool3d
				}.get(len(self.size))
			elif self.pooltype == 'average':
				func = {
					1 : nn.AvgPool1d,
					2 : nn.AvgPool2d,
					3 : nn.AvgPool3d
				}.get(len(self.size))
			else:
				raise ValueError('Unhandled pool type "{}". This is a bug.',
					self.pooltype)

			def connect(inputs):
				""" Connects the layers.
				"""
				assert len(inputs) == 1
				output = model.data.add_operation(
					swap_channels
				)(inputs[0]['layer'])
				output = model.data.add_layer(
					self.name,
					func(
						self.size,
						self.strides,
						padding=0,
						dilation=1,
						ceil_mode=False
					)
				)(output)
				output = model.data.add_operation(
					swap_channels
				)(output)
				return {
					'shape' : self.shape([inputs[0]['shape']]),
					'layer' : output
				}

			yield connect

		else:
			raise ValueError(
				'Unknown or unsupported backend: {}'.format(backend))

	###########################################################################
	def shape(self, input_shapes):
		""" Returns the output shape of this layer for a given input shape.
		"""
		if len(input_shapes) > 1:
			raise ValueError('Pooling layers only take a single input.')
		input_shape = input_shapes[0]
		if len(input_shape) != len(self.size) + 1:
			raise ValueError('Invalid input shape to a pooling layer: {}'
				.format(input_shape))

		output_shape = tuple(
			(input_shape[i] - self.size[i]) // self.strides[i] + 1
			for i in range(len(self.size))
		) + (input_shape[-1], )
		return output_shape

### EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF



import linecache
import numpy
import json

from ..sources import ChunkSource, DerivedSource


###############################################################################
class TextLength(DerivedSource):
    """ Data source for audio lengths.
    """
    def __init__(self, source, num_entries):
        super().__init__()
        self.source = source
        self.num_entries = num_entries
    def derive(self, inputs):
        text_chunks, = inputs
        return numpy.array([[len(x)] for x in text_chunks], dtype='int32')
    def shape(self):
        return (1,)
    def requires(self):
        return (self.source, )
    def __len__(self):
        return self.num_entries

###############################################################################
class TextSource(DerivedSource):
    """ Data source for neat (non-ragged) one-hot represented text arrays.
    """
    def __init__(self, source, vocab, num_entries, seq_len, padding='right', pad_with=None):
        super().__init__()
        self.num_entries = num_entries
        self.source = source
        self.seq_len = seq_len
        self.padding = padding
        self.pad_with = pad_with
        self.vocab = vocab
        self.char_to_int = {
            c: i
            for i, c in enumerate(self.vocab)
        }

    def _encode(self, char_seq):
        output = numpy.zeros(shape=(len(char_seq), len(self.vocab),))
        for i in range(len(char_seq)):
            output[i][self.char_to_int[char_seq[i]]] = 1
        assert output.shape[0] == len(char_seq)
        return output

    def derive(self, inputs):
        text_chunk, = inputs

        output = numpy.zeros(
            shape=(
                len(text_chunk),
                self.seq_len,
                len(self.vocab),
            ),
            dtype='int32'
        )

        for i, char_seq in enumerate(text_chunk):
            char_seq = list(char_seq)
            if self.padding == 'right':
                if self.pad_with is not None:
                    char_seq = char_seq + (self.seq_len - len(char_seq)) * [self.pad_with]
                encoded_char_seq = self._encode(char_seq)

                assert len(encoded_char_seq) == len(char_seq)

                for j in range(len(encoded_char_seq)):
                    output[i][j] = encoded_char_seq[j]

            elif self.padding == 'left':
                if self.pad_with is not None:
                    char_seq = (self.seq_len - len(char_seq)) * [self.pad_with] + char_seq
                encoded_char_seq = self._encode(char_seq)

                assert len(encoded_char_seq) == len(char_seq)

                for j in range(len(encoded_char_seq)):
                    output[i][-len(char_seq) + j] = encoded_char_seq[j]

            else:
                raise ValueError('Padding must be left or right, not %s' % padding)
        return output

    def shape(self):
        """ Return the shape of the tensor (excluding batch size) returned by
            this data source.
        """
        return (self.seq_len, len(self.vocab),)

    def requires(self):
        return (self.source, )

    def __len__(self):
        return self.num_entries

###############################################################################
class RawText(ChunkSource):
    """ Data source for text stored in JSONL format like:
    ['a', 'p', 'p', 'l', 'e', ' ', 'p', 'i', 'e']
    """

    ###########################################################################
    @classmethod
    def default_chunk_size(cls):
        """ Returns the default chunk size for this source.
        """
        return 256

    ########################################################################### 
    def shape(self):
        return (None,)

    ###########################################################################
    def __init__(self, source, key, num_entries, *args, **kwargs):
        """ Creates a new Text source for file named `source`.
        """

        super().__init__(*args, **kwargs)

        self.source = source
        self.num_entries = num_entries
        self.key = key
        self.indices = numpy.arange(len(self))

    ###########################################################################
    def __iter__(self):
        """ Return an iterator to the data. Yield the value for self.key
        from each object
        """
        start = 0
        while start < self.num_entries:
            end = min(self.num_entries, start + self.chunk_size)

            # linecache line numbering starts at 1
            batch = [
                json.loads(
                    linecache.getline(
                        self.source,
                        i + 1
                    ).strip()
                )[self.key]
                for i in self.indices[start:end]
            ]

            yield batch
            start = end

    ###########################################################################
    def __len__(self):
        """ Returns the total number of entries that this source can return, if
            known.
        """
        return self.num_entries

    ###########################################################################
    def can_shuffle(self):
        """ This source can be shuffled.
        """
        return True

    ###########################################################################
    def shuffle(self, indices):
        """ Applies a permutation to the data.
        """
        if len(indices) > len(self):
            raise ValueError('Shuffleable was asked to apply permutation, but '
                'the permutation is longer than the length of the data set.')
        self.indices[:len(indices)] = self.indices[:len(indices)][indices]



"""
Copyright 2016 Deepgram

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import os
import copy
import socket
import warnings
import logging
from collections import deque

from .engine import ScopeStack, PassthroughEngine
from .reader import Reader
from .containers import Container, ParsingError
from .model import Model, Executor, EvaluationHook, OutputHook, TrainingHook
from .backend import Backend
from .optimizer import Optimizer, Adam
from .loss import Loss
from .providers import Provider, BatchProvider, ShuffleProvider
from .supplier import Supplier
from .utils import mergetools
from .loggers import Logger

logger = logging.getLogger(__name__)

###############################################################################
class Kurfile:
	""" Class for loading and parsing Kurfiles.
	"""

	DEFAULT_OPTIMIZER = Adam
	DEFAULT_PROVIDER = BatchProvider

	###########################################################################
	def __init__(self, source, engine=None):
		""" Creates a new Kurfile.

			# Arguments

			source: str or dict. If it is a string, it is interpretted as a
				filename to an on-disk Kurfile. Otherwise, it is
				interpretted as an already-loaded, but unparsed, Kurfile.
			engine: Engine instance. The templating engine to use in parsing.
				If None, a Passthrough engine is instantiated.
		"""
		engine = engine or PassthroughEngine()
		if isinstance(source, str):
			filename = os.path.expanduser(os.path.expandvars(source))
			if not os.path.isfile(filename):
				raise IOError('No such file found: {}. Path was expanded to: '
					'{}.'.format(source, filename))
			self.filename = filename
			self.data = self.parse_source(
				engine,
				source=filename,
				context=None
			)
		else:
			self.filename = None
			self.data = dict(source)

		self.containers = None
		self.model = None
		self.backend = None
		self.engine = engine

	###########################################################################
	def parse(self):
		""" Parses the Kurfile.
		"""

		logger.info('Parsing Kurfile...')

		# These are reserved section names.
		# The first name in each tuple is the one that we should rely on
		# internally when we reference `self.data` in Kurfile.
		builtin = {
			'settings' : ('settings', ),
			'train' : ('train', 'training'),
			'validate' : ('validate', 'validation'),
			'test' : ('test', 'testing'),
			'evaluate' : ('evaluate', 'evaluation'),
			'templates' : ('templates', ),
			'model' : ('model', ),
			'loss' : ('loss', )
		}

		# The scope stack.
		stack = []

		# Add default entries.
		stack.append({
			'filename' : self.filename,
			'raw' : copy.deepcopy(self.data),
			'parsed' : self.data,
			'host' : socket.gethostname()
		})

		# Parse the settings (backends, globals, hyperparameters, ...)
		self._parse_section(
			self.engine, builtin['settings'], stack, include_key=False)

		# Parse all of the "action" sections.
		self._parse_section(
			self.engine, builtin['train'], stack, include_key=True)
		self._parse_section(
			self.engine, builtin['validate'], stack, include_key=True)
		self._parse_section(
			self.engine, builtin['test'], stack, include_key=True)
		self._parse_section(
			self.engine, builtin['evaluate'], stack, include_key=True)

		# Parse the templates
		self.templates = self._parse_templates(
			self.engine, builtin['templates'], stack)

		# Parse the model.
		self.containers = self._parse_model(
			self.engine, builtin['model'], stack, required=True)

		# Parse the loss function.
		self._parse_section(
			self.engine, builtin['loss'], stack, include_key=True)

		# Check for unused keys.
		for key in self.data.keys():
			for section in builtin.values():
				if key in section:
					break
			else:
				warnings.warn('Unexpected section in Kurfile: "{}". '
					'This section will be ignored.', SyntaxWarning)

	###########################################################################
	def get_model(self, provider=None):
		""" Returns the parsed Model instance.
		"""
		if self.model is None:
			if self.containers is None:
				raise ValueError('No such model available.')
			self.model = Model(
				backend=self.get_backend(),
				containers=self.containers
			)
			self.model.parse(self.engine)
			self.model.register_provider(provider)
			self.model.build()
		return self.model

	###########################################################################
	def get_backend(self):
		""" Creates a new backend from the Kurfile.

			# Return value

			Backend instance

			# Notes

			- If no "global" section exists, or if no "backend" section exists
			  in the "global" section, this will try to instantiate any
			  supported backend installed on the system.
			- This passes `self.data['global']['backend']` to the Backend
			  factory method, if such keys exist; otherwise, it passes None to
			  the factory method.
		"""
		if self.backend is None:
			self.backend = Backend.from_specification(
				(self.data.get('settings') or {}).get('backend')
			)
		return self.backend

	###########################################################################
	def get_seed(self):
		""" Retrieves the global random seed, if any.

			# Return value

			The `settings.seed` value, as an integer, if such a key exists.
			If the key does not exist, returns None.
		"""

		if 'settings' not in self.data or \
				not isinstance(self.data['settings'], dict):
			return None

		seed = self.data['settings'].get('seed')
		if seed is None:
			return None
		elif not isinstance(seed, int):
			raise ValueError('Unknown format for random seed: {}'.format(seed))

		if seed < 0 or seed >= 2**32:
			logger.warning('Random seeds should be unsigned, 32-bit integers. '
				'We are truncating the seed.')
			seed %= 2**32

		return seed

	###########################################################################
	def get_provider(self, section):
		""" Creates the provider corresponding to a part of the Kurfile.

			# Arguments

			section: str. The name of the section to load the provider for.

			# Return value

			If the section exists and has at least the "data" or "provider"
			section defined, then this returns a Provider instance. Otherwise,
			returns None.
		"""
		if section in self.data:
			section = self.data[section]
			if not any(k in section for k in ('data', 'provider')):
				return None
		else:
			return None

		supplier_list = section.get('data') or []
		if not isinstance(supplier_list, (list, tuple)):
			raise ValueError('"data" section should be a list.')
		suppliers = [
			Supplier.from_specification(x, kurfile=self) for x in supplier_list
		]

		provider_spec = dict(section.get('provider') or {})
		if 'name' in provider_spec:
			provider = Provider.get_provider_by_name(provider_spec.pop('name'))
		else:
			provider = Kurfile.DEFAULT_PROVIDER

		return provider(
			sources=Supplier.merge_suppliers(suppliers),
			**provider_spec
		)

	###########################################################################
	def get_training_function(self):
		""" Returns a function that will train the model.
		"""
		if 'train' not in self.data:
			raise ValueError('Cannot construct training function. There is a '
				'missing "train" section.')

		if 'log' in self.data['train']:
			log = Logger.from_specification(self.data['train']['log'])
		else:
			log = None

		epochs = self.data['train'].get('epochs')

		provider = self.get_provider('train')

		training_hooks = self.data['train'].get('hooks', [])
		if not isinstance(training_hooks, (list, tuple)):
			raise ValueError('"hooks" (in the "train" section) should '
				'be a list of hook specifications.')
		training_hooks = [TrainingHook.from_specification(spec) \
			for spec in training_hooks]

		if 'validate' in self.data:
			validation = self.get_provider('validate')
			validation_weights = self.data['validate'].get('weights')
			if validation_weights is None:
				best_valid = None
			elif isinstance(validation_weights, str):
				best_valid = validation_weights
			elif isinstance(validation_weights, dict):
				best_valid = validation_weights.get('best')
			else:
				raise ValueError('Unknown type for validation weights: {}'
					.format(validation_weights))

			validation_hooks = self.data['validate'].get('hooks', [])
			if not isinstance(validation_hooks, (list, tuple)):
				raise ValueError('"hooks" (in the "validate" section) should '
					'be a list of hook specifications.')
			validation_hooks = [EvaluationHook.from_specification(spec) \
				for spec in validation_hooks]
		else:
			validation = None
			best_valid = None
			validation_hooks = None

		train_weights = self.data['train'].get('weights')
		if train_weights is None:
			initial_weights = best_train = last_weights = None
			deprecated_checkpoint = None
		elif isinstance(train_weights, str):
			initial_weights = train_weights
			best_train = train_weights if best_valid is None else None
			last_weights = None
			initial_must_exist = False
			deprecated_checkpoint = None
		elif isinstance(train_weights, dict):
			initial_weights = train_weights.get('initial')
			best_train = train_weights.get('best')
			last_weights = train_weights.get('last')
			initial_must_exist = train_weights.get('must_exist', False)
			deprecated_checkpoint = train_weights.get('checkpoint')
		else:
			raise ValueError('Unknown weight specification for training: {}'
				.format(train_weights))

		checkpoint = self.data['train'].get('checkpoint')

		if deprecated_checkpoint is not None:
			warnings.warn('"checkpoint" belongs under "train", not under '
				'"weights".', DeprecationWarning)
			if checkpoint is None:
				checkpoint = deprecated_checkpoint
			else:
				logger.warning('The currently-accepted "checkpoint" will be '
					'used over the deprecated "checkpoint".')

		expand = lambda x: os.path.expanduser(os.path.expandvars(x))
		initial_weights, best_train, best_valid, last_weights = [
			expand(x) if x is not None else x for x in
				(initial_weights, best_train, best_valid, last_weights)
		]

		model = self.get_model(provider)
		trainer = self.get_trainer()

		def func(**kwargs):
			""" Trains a model from a pre-packaged specification file.
			"""
			if initial_weights is not None:
				if os.path.exists(initial_weights):
					model.restore(initial_weights)
				elif initial_must_exist:
					logger.error('Configuration indicates that the weight '
						'file must exist, but the weight file was not found.')
					raise ValueError('Missing initial weight file. If you '
						'want to proceed anyway, set "must_exist" to "no" '
						'under the training "weights" section.')
				else:
					if log is not None and \
						(log.get_number_of_epochs() or 0) > 0:
						logger.warning('The initial weights are missing: %s. '
							'However, the logs suggest that this model has '
							'been trained previously. It is possible that the '
							'log file is corrupt/out-of-date, or that the '
							'weight files are not supposed to be missing.',
							initial_weights)
					else:
						logger.info('Ignoring missing initial weights: %s. If '
							'this is undesireable, set "must_exist" to "yes" '
							'in the approriate "weights" section.',
							initial_weights)
			defaults = {
				'provider' : provider,
				'validation' : validation,
				'epochs' : epochs,
				'log' : log,
				'best_train' : best_train,
				'best_valid' : best_valid,
				'last_weights' : last_weights,
				'training_hooks' : training_hooks,
				'validation_hooks' : validation_hooks,
				'checkpoint' : checkpoint
			}
			defaults.update(kwargs)
			return trainer.train(**defaults)

		return func

	###########################################################################
	def get_testing_function(self):
		""" Returns a function that will test the model.
		"""
		if 'test' not in self.data:
			raise ValueError('Cannot construct testing function. There is a '
				'missing "test" section.')

		provider = self.get_provider('test')

		# No reason to shuffle things for this.
		if isinstance(provider, ShuffleProvider):
			provider.randomize = False

		weights = self.data['test'].get('weights')
		if weights is None:
			initial_weights = None
		elif isinstance(weights, str):
			initial_weights = weights
		elif isinstance(weights, dict):
			initial_weights = weights.get('initial')
		else:
			raise ValueError('Unknown weight specification for testing: {}'
				.format(weights))

		hooks = self.data['test'].get('hooks', [])
		if not isinstance(hooks, (list, tuple)):
			raise ValueError('"hooks" (in the "test" section) should be a '
			   'list of hook specifications.')
		hooks = [EvaluationHook.from_specification(spec) for spec in hooks]

		expand = lambda x: os.path.expanduser(os.path.expandvars(x))
		if initial_weights is not None:
			initial_weights = expand(initial_weights)

		model = self.get_model(provider)
		trainer = self.get_trainer(with_optimizer=False)

		def func(**kwargs):
			""" Tests a model from a pre-packaged specification file.
			"""
			if initial_weights is not None:
				if os.path.exists(initial_weights):
					model.restore(initial_weights)
				else:
					logger.error('No weight file found: %s. We will just '
						'proceed with default-initialized weights. This will '
						'test that the system works, but the results will be '
						'terrible.', initial_weights)
			else:
				logger.warning('No weight file specified. We will just '
					'proceed with default-initialized weights. This will '
					'test that the system works, but the results will be '
					'terrible.')
			defaults = {
				'provider' : provider,
				'validating' : False,
				'hooks' : hooks
			}
			defaults.update(kwargs)
			return trainer.test(**defaults)

		return func

	###########################################################################
	def get_trainer(self, with_optimizer=True):
		""" Creates a new Trainer from the Kurfile.

			# Return value

			Trainer instance.
		"""
		return Executor(
			model=self.get_model(),
			loss=self.get_loss(),
			optimizer=self.get_optimizer() if with_optimizer else None
		)

	###########################################################################
	def get_optimizer(self):
		""" Creates a new Optimizer from the Kurfile.

			# Return value

			Optimizer instance.
		"""
		if 'train' not in self.data:
			raise ValueError('Cannot construct optimizer. There is a missing '
				'"train" section.')

		spec = dict(self.data['train'].get('optimizer', {}))
		if 'name' in spec:
			optimizer = Optimizer.get_optimizer_by_name(spec.pop('name'))
		else:
			optimizer = Kurfile.DEFAULT_OPTIMIZER

		return optimizer(**spec)

	###########################################################################
	def get_loss(self):
		""" Creates a new Loss from the Kurfile.

			# Return value

			If 'loss' is defined, then this returns a dictionary whose keys are
			output layer names and whose respective values are Loss instances.
			Otherwise, this returns None.
		"""
		if 'loss' not in self.data:
			return None

		spec = self.data['loss']
		if not isinstance(spec, (list, tuple)):
			raise ValueError('"loss" section expects a list of loss '
				'functions.')

		result = {}
		for entry in spec:
			entry = dict(entry)
			target = entry.pop('target', None)
			if target is None:
				raise ValueError('Each loss function in the "loss" list '
					'must have a "target" entry which names the output layer '
					'it is attached to.')

			name = entry.pop('name', None)
			if name is None:
				raise ValueError('Each loss function in the "loss" list '
					'must have a "name" entry which names the loss function '
					'to use for that output.')

			if target in result:
				logger.warning('A loss function for the "%s" output was '
					'already defined. We will replace it by this new one we '
					'just found.', target)
			result[target] = Loss.get_loss_by_name(name)(**entry)

		return result

	###########################################################################
	def get_evaluation_function(self):
		""" Returns a function that will evaluate the model.
		"""
		if 'evaluate' not in self.data:
			raise ValueError('Cannot construct evaluation function. There is '
				'a missing "evaluate" section.')

		provider = self.get_provider('evaluate')

		# No reason to shuffle things for this.
		if isinstance(provider, ShuffleProvider):
			provider.randomize = False

		weights = self.data['evaluate'].get('weights')
		if weights is None:
			initial_weights = None
		elif isinstance(weights, str):
			initial_weights = weights
		elif isinstance(weights, dict):
			initial_weights = weights.get('initial')
		else:
			raise ValueError('Unknown weight specification for evaluation: {}'
				.format(weights))

		destination = self.data['evaluate'].get('destination')
		if isinstance(destination, str):
			destination = OutputHook(path=destination)
		elif isinstance(destination, dict):
			destination = OutputHook(**destination)
		elif destination is not None:
			ValueError('Expected a string or dictionary value for '
				'"destination" in "evaluate". Received: {}'.format(
					destination))

		hooks = self.data['evaluate'].get('hooks', [])
		if not isinstance(hooks, (list, tuple)):
			raise ValueError('"hooks" should be a list of hook '
				'specifications.')
		hooks = [EvaluationHook.from_specification(spec) for spec in hooks]

		expand = lambda x: os.path.expanduser(os.path.expandvars(x))
		if initial_weights is not None:
			initial_weights = expand(initial_weights)

		model = self.get_model(provider)
		evaluator = self.get_evaluator()

		def func(**kwargs):
			""" Evaluates a model from a pre-packaged specification file.
			"""
			if initial_weights is not None:
				if os.path.exists(initial_weights):
					model.restore(initial_weights)
				else:
					logger.error('No weight file found: %s. We will just '
						'proceed with default-initialized weights. This will '
						'test that the system works, but the results will be '
						'terrible.', initial_weights)
			else:
				logger.warning('No weight file specified. We will just '
					'proceed with default-initialized weights. This will '
					'test that the system works, but the results will be '
					'terrible.')
			defaults = {
				'provider' : provider,
				'callback' : None
			}
			defaults.update(kwargs)
			result, truth = evaluator.evaluate(**defaults)
			orig = (result, truth)
			prev = orig
			for hook in hooks:
				new_prev = hook.apply(prev, orig, model)
				prev = (new_prev, prev[1]) \
					if not isinstance(new_prev, tuple) else new_prev
			if destination is not None:
				new_prev = destination.apply(prev, orig, model)
				prev = (new_prev, prev[1]) \
					if not isinstance(new_prev, tuple) else new_prev
			return prev

		return func

	###########################################################################
	def get_evaluator(self):
		""" Creates a new Evaluator from the Kurfile.

			# Return value

			Evaluator instance.
		"""
		return Executor(
			model=self.get_model()
		)

	###########################################################################
	def _parse_model(self, engine, section, stack, *, required=True):
		""" Parses the top-level "model" entry.

			# Arguments

			engine: Engine instance. The templating engine to use in
				evaluation.
			section: str or list of str. The key indicating which top-level
				entry to parse. If a list, the first matching key is used; this
				allows for multiple aliases to the same section.
			stack: list. The list of scope dictionaries to use in evaluation.
			required: bool (default: True). If True, an exception is raised if
				the key is not present in the Kurfile.

			# Return value

			If the model section is found, the list of containers is returned.
			If the model section is not found but is required, an exception is
			raised. If the model section is not found and isn't required, None
			is returned.
		"""
		if isinstance(section, str):
			section = (section, )

		key = None		# Not required, but shuts pylint up.
		for key in section:
			if key in self.data:
				break
		else:
			if required:
				raise ValueError(
					'Missing required section: {}'.format(', '.join(section)))
			else:
				return None

		if not isinstance(self.data[key], (list, tuple)):
			raise ValueError(
				'Section "{}" should contain a list of layers.'.format(key))

		with ScopeStack(engine, stack):
			queue = deque(self.data[key])
			containers = []
			while queue:
				entry = queue.popleft()
				entry = engine.evaluate(entry)
				containers.append(
					Container.create_container_from_data(entry)
				)
			for container in containers:
				container.parse(engine)

		return containers

	###########################################################################
	def parse_source(self, engine, *,
		source=None, context=None, loaded=None):
		""" Parses a source, and its includes (recursively), and returns the
			merged sources.
		"""
		def get_id(filename):
			""" Returns a tuple which uniquely identifies a file.

				This function follows symlinks.
			"""
			filename = os.path.expanduser(os.path.expandvars(filename))
			stat = os.stat(filename)
			return (stat.st_dev, stat.st_ino)

		# Process the "loaded" iterable.
		loaded = loaded or set()

		if isinstance(source, str):
			filename = source
			strategy = None
		elif isinstance(source, dict):
			if 'source' not in source:
				raise ParsingError('Error while parsing source file {}. '
					'Missing required "source" key in the include '
					'dictionary: {}.'.format(context or 'top-level', source))
			filename = source.pop('source')
			strategy = source.pop('method', None)
			for k in source:
				warnings.warn('Warning while parsing source file {}. '
					'Ignoring extra key in an "include" dictionary: {}.'
					.format(context or 'top-level', k))
		else:
			raise ParsingError('Error while parsing source file {}. '
				'Expected each "include" to be a string or a dictionary. '
				'Received: {}'.format(context or 'top-level', source))

		logger.info('Parsing source: %s, included by %s.', source,
			context or 'top-level')

		if context:
			filename = os.path.join(os.path.dirname(context), filename)
		expanded = os.path.expanduser(os.path.expandvars(filename))
		if not os.path.isfile(expanded):
			raise IOError('Error while parsing source file {}. No such '
				'source file found: {}. Path was expanded to: {}'.format(
				context or 'top-level', filename, expanded))
		file_id = get_id(expanded)
		if file_id in loaded:
			logger.warning('Skipping an already included source file: %s. '
				'This may have unintended consequences, since the merge '
				'order may result in different final results. You should '
				'refactor your Kurfile to avoid circular includes.',
				expanded)
			return {}
		else:
			loaded.add(file_id)

		try:
			data = Reader.read_file(expanded)
		except:
			logger.exception('Failed to read file: %s. Check your syntax.',
				expanded)
			raise

		new_sources = data.pop('include', [])
		engine.evaluate(new_sources, recursive=True)
		if isinstance(new_sources, str):
			new_sources = [new_sources]
		elif not isinstance(new_sources, (list, tuple)):
			raise ValueError('Error while parsing file: {}. All "include" '
				'sections must be a single source file or a list of source '
				'files. Instead we received: {}'.format(
					expanded, new_sources))

		for x in new_sources:
			data = mergetools.deep_merge(
				self.parse_source(
					engine,
					source=x,
					context=expanded,
					loaded=loaded
				),
				data,
				strategy=strategy
			)

		return data

	###########################################################################
	def _parse_templates(self, engine, section, stack):
		""" Parses the template section.
		"""
		if isinstance(section, str):
			section = (section, )

		key = None		# Not required, but shuts pylint up.
		for key in section:
			if key in self.data:
				break
		else:
			return None

		if not self.data[key]:
			return None

		if not isinstance(self.data[key], dict):
			raise ValueError('Section "{}" should contain a dictionary of '
				'templates.'.format(key))

		for template in self.data[key]:
			other = Container.get_container_for_name(template)
			if other:
				raise ValueError('Templates cannot override built-in layers. '
					'We found a conflict with template "{}".'.format(template))

		engine.register_templates(self.data[key])

		return self.data[key]

	###########################################################################
	def _parse_section(self, engine, section, stack, *,
		required=False, include_key=True):
		""" Parses a single top-level entry in the Kurfile.

			# Arguments

			engine: Engine instance. The templating engine to use in
				evaluation.
			section: str or list of str. The key indicating which top-level
				entry to parse. If a list, the first matching key is used; this
				allows for multiple aliases to the same section.
			stack: list. The list of scope dictionaries to use in evaluation.
			required: bool (default: False). If True, an exception is raised if
				the key is not present in the Kurfile.
			include_key: bool (default: False). If True, the parsed section is
				added to the scope stack as a dictionary with a single item,
				whose key is the name of the parsed section, and whose value is
				the evaluated section. If False, then the evaluated section is
				added directly to the scope stack (it must evaluate to a
				dictionary for this to work).

			# Return value

			The parsed section.

			# Note

			- This will replace the values of `self.data` with the evaluated
			  version.
		"""
		if isinstance(section, str):
			section = (section, )

		logger.debug('Parsing Kurfile section: %s', section[0])

		key = None		# Not required, but shuts pylint up.
		for key in section:
			if key in self.data:
				break
		else:
			if required:
				raise ValueError(
					'Missing required section: {}'.format(', '.join(section)))
			else:
				return None

		with ScopeStack(engine, stack + [{key : self.data[key]}]):
			evaluated = engine.evaluate(self.data[key], recursive=True)

		if not include_key:
			if evaluated is not None:
				if not isinstance(evaluated, dict):
					raise ValueError(
						'Section "{}" should contain key/value pairs.'
						.format(key))
				stack.append(evaluated)
		stack.append({key : evaluated})

		# Now, just in case this was aliased, let's make sure we keep our
		# naming scheme consistent.
		for key in section:
			self.data[key] = evaluated

		return evaluated

### EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF



"""
Copyright 2017 Deepgram

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from . import Optimizer, keras_clip, keras_wrap

###############################################################################
class RMSProp(Optimizer):
	""" RMSProp optimizer
	"""

	###########################################################################
	def __init__(self, learning_rate=None, rho=None, epsilon=None, decay=None,
		*args, **kwargs):
		""" Create a new RMSProp optimizer.

			# Arguments

			learning_rate: float. The learning rate to use.
		"""
		super().__init__(*args, **kwargs)

		self.learning_rate = learning_rate or 0.001
		self.rho = rho or 0.9
		self.epsilon = epsilon or 1e-8
		self.decay = decay or 0.0

		self.optimizer = None

	###########################################################################
	def get_optimizer(self, backend):
		""" Returns a backend-specific instantiation of the optimizer.
		"""
		if backend.get_name() == 'keras':
			import keras.optimizers as O		# pylint: disable=import-error
			self.optimizer = self.optimizer or O.RMSprop(
				lr=self.learning_rate,
				rho=self.rho,
				epsilon=self.epsilon,
				decay=self.decay,
				**keras_clip(self)
			)
			return keras_wrap(self.optimizer)
		else:
			raise ValueError('Unsupported backend "{}" for optimizer "{}"'
				.format(backend.get_name(), self.get_name()))

### EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF



"""
Copyright 2017 Anthony Rousseau for Allo-media
Based on kuza55 keras-extras (https://github.com/kuza55/keras-extras)
and jonilaserson (https://github.com/fchollet/keras/issues/2436) code.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import keras.layers as L
from keras.layers.core import Lambda
from keras import backend as K

import tensorflow as tf
from ..backend import KerasBackend

def make_parallel(model, gpu_count):
	""" Allows a model to run on multiple GPUs.
	"""
	def slice_batch(x, n_gpus, part):
		""" Divide the input batch into [n_gpus] slices,
			and obtain slice no. [part].
			i.e. if len(x) = 10, then slice_batch(x, 2, 1)
			will return x[5:].
		"""
		sh = K.shape(x)
		L = sh[0] // n_gpus # sh[0] = batch_size

		if part == n_gpus - 1:
			return x[part*L:]

		return x[part*L:(part+1)*L]

	all_outputs = []

	# Empty list for each output in the model
	for i in range(len(model.outputs)):
		all_outputs.append([])

	# Place a copy of the model on each GPU, 
	# each getting a slice of the batch
	for i in range(gpu_count):
		with tf.device('/gpu:%d' % i):
			with tf.name_scope('tower_%d' % i) as scope:
				inputs = []

				# Slice each input into a piece 
				# for processing on this GPU
				for x in model.inputs:
					input_shape = (None,) + tuple(x.get_shape().as_list())[1:]
					slice_n = Lambda(slice_batch, 
						lambda shape: input_shape, 
						arguments={'n_gpus':gpu_count, 'part':i})(x)
					inputs.append(slice_n)

				outputs = model(inputs)

				if not isinstance(outputs, list):
					outputs = [outputs]

				# Save all the outputs for 
				# merging back together later
				for l in range(len(outputs)):
					all_outputs[l].append(outputs[l])

	# Merge outputs on CPU
	with tf.device('/cpu:0'):
		merged = []

		if KerasBackend.keras_version() == 1:
			func = lambda x: L.merge(x, mode='concat', concat_axis=0)
		else:
			func = lambda x: L.concatenate(x, axis=0)

		for outputs in all_outputs:
			merged.append(func(outputs))

	return KerasBackend.make_model(inputs=model.inputs, outputs=merged)



"""
Copyright 2016 Deepgram

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import pytest

from kur.loss import MeanSquaredError, Ctc
from kur.optimizer import Adam
from kur.model import Executor
from kur.providers import BatchProvider

Executor.DEFAULT_RETRY_ENABLED = False

###############################################################################
@pytest.fixture
def ctc_eval_data(ctc_data):
	""" Provides CTC data with only the evaluation-time input data available.
	"""
	for key, source in zip(ctc_data.keys, ctc_data.sources):
		if key == 'TEST_input':
			return BatchProvider(sources={key : source})

@pytest.fixture
def loss():
	return MeanSquaredError()

@pytest.fixture
def ctc_loss():
	return Ctc(
		input_length='TEST_input_length',
		output_length='TEST_transcription_length',
		output='TEST_transcription'
	)

@pytest.fixture
def optimizer():
	return Adam()

###############################################################################
class TestExecutor:
	""" Tests for the Executor (training/validation/testing).
	"""

	###########################################################################
	def test_trainer_with_optimizer(self, simple_model, loss, optimizer):
		""" Tests if we can compile a Executor instance with an optimizer.
		"""
		simple_model.parse(None)
		simple_model.build()
		trainer = Executor(
			model=simple_model,
			loss=loss,
			optimizer=optimizer
		)
		trainer.compile()

	###########################################################################
	def test_trainer_without_optimizer(self, simple_model, loss):
		""" Tests if we can compile a Executor instance without an optimizer.
		"""
		simple_model.parse(None)
		simple_model.build()
		trainer = Executor(
			model=simple_model,
			loss=loss
		)
		trainer.compile()

	###########################################################################
	def test_training(self, simple_model, loss, optimizer, simple_data):
		""" Tests that a single training epoch can succeed.
		"""
		simple_model.parse(None)
		simple_model.build()
		trainer = Executor(
			model=simple_model,
			loss=loss,
			optimizer=optimizer
		)
		trainer.train(provider=simple_data, epochs=1)

	###########################################################################
	def test_testing(self, simple_model, loss, simple_data):
		""" Tests that a testing run can succeed.
		"""
		simple_model.parse(None)
		simple_model.build()
		trainer = Executor(
			model=simple_model,
			loss=loss
		)
		trainer.test(provider=simple_data)

	###########################################################################
	def test_evaluator(self, simple_model):
		""" Tests if we can compile an Executor instance.
		"""
		simple_model.parse(None)
		simple_model.build()
		evaluator = Executor(
			model=simple_model
		)
		evaluator.compile()

	###########################################################################
	def test_evaluating(self, simple_model, simple_data):
		""" Tests that an evaluation run can succeed.
		"""
		simple_model.parse(None)
		simple_model.build()
		evaluator = Executor(
			model=simple_model
		)
		evaluator.evaluate(provider=simple_data)

	###########################################################################
	def test_ctc_train(self, ctc_model, ctc_data, ctc_loss, optimizer):
		""" Tests that we can compile and train a model using the CTC loss
			function.
		"""
		ctc_model.parse(None)
		ctc_model.register_provider(ctc_data)
		ctc_model.build()
		trainer = Executor(
			model=ctc_model,
			loss=ctc_loss,
			optimizer=optimizer
		)
		trainer.train(provider=ctc_data, epochs=1)

	###########################################################################
	def test_ctc_test(self, ctc_model, ctc_data, ctc_loss):
		""" Tests that we can compile and test a model using the CTC loss
			function.
		"""
		ctc_model.parse(None)
		ctc_model.register_provider(ctc_data)
		ctc_model.build()
		trainer = Executor(
			model=ctc_model,
			loss=ctc_loss
		)
		trainer.test(provider=ctc_data)

	###########################################################################
	def test_ctc_evaluating(self, ctc_model, ctc_eval_data):
		""" Tests that we can evaluate a model that was trained with CTC loss.
		"""
		ctc_model.parse(None)
		ctc_model.register_provider(ctc_eval_data)
		ctc_model.build()
		evaluator = Executor(
			model=ctc_model
		)
		evaluator.evaluate(provider=ctc_eval_data)

	###########################################################################
	def test_uber_train(self, uber_model, uber_data, jinja_engine, loss,
		optimizer):
		""" Tests that we can compile and train a diverse model.
		"""
		uber_model.parse(jinja_engine)
		uber_model.register_provider(uber_data)
		uber_model.build()
		trainer = Executor(
			model=uber_model,
			loss=loss,
			optimizer=optimizer
		)
		trainer.compile()
		trainer.train(provider=uber_data, epochs=1)

	###########################################################################
	def test_uber_test(self, uber_model, uber_data, jinja_engine, loss):
		""" Tests that we can compile and test a diverse model.
		"""
		uber_model.parse(jinja_engine)
		uber_model.register_provider(uber_data)
		uber_model.build()
		trainer = Executor(
			model=uber_model,
			loss=loss
		)
		trainer.compile()
		trainer.test(provider=uber_data)

	###########################################################################
	def test_uber_evaluating(self, uber_model, uber_data, jinja_engine):
		""" Tests that we can evaluate a very diverse model.
		"""
		uber_model.parse(jinja_engine)
		uber_model.register_provider(uber_data)
		uber_model.build()
		evaluator = Executor(
			model=uber_model
		)
		evaluator.evaluate(provider=uber_data)

	###########################################################################
	def test_embedding_train(self, embedding_model, embedding_data, loss,
		optimizer):
		""" Tests that we can compile and train a model which has an Embedding.
		"""
		embedding_model.parse(None)
		embedding_model.register_provider(embedding_data)
		embedding_model.build()
		trainer = Executor(
			model=embedding_model,
			loss=loss,
			optimizer=optimizer
		)
		trainer.train(provider=embedding_data, epochs=1)

	###########################################################################
	def test_embedding_test(self, embedding_model, embedding_data, loss):
		""" Tests that we can compile and test a model which has an Embedding.
			function.
		"""
		embedding_model.parse(None)
		embedding_model.register_provider(embedding_data)
		embedding_model.build()
		trainer = Executor(
			model=embedding_model,
			loss=loss
		)
		trainer.test(provider=embedding_data)

	###########################################################################
	def test_embedding_evaluating(self, embedding_model, embedding_data):
		""" Tests that we can evaluate a model that has an Embedding.
		"""
		embedding_model.parse(None)
		embedding_model.register_provider(embedding_data)
		embedding_model.build()
		evaluator = Executor(
			model=embedding_model
		)
		evaluator.evaluate(provider=embedding_data)

### EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF



"""
Copyright 2016 Deepgram

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import logging
import types
from collections import namedtuple, OrderedDict, deque
from ..containers import Container
from ..containers.operators import ContainerGroup
from ..containers.layers import Placeholder
from ..engine import PassthroughEngine

logger = logging.getLogger(__name__)

# Convenience class for keeping track of high-level network nodes.
ContainerNode = types.SimpleNamespace

CollapsedContainer = namedtuple('CollapsedContainer',
	['inputs', 'container', 'names']
)

###############################################################################
class Model:
	""" Class for holding the model corresponding to a given container.

		Unlike a container, a model's layers are connected vis-a-vis the
		backend, and it has tracked input and output connections.
	"""

	###########################################################################
	def __init__(self, backend, containers):
		""" Create a new model.

			# Arguments

			backend: Backend instance. The backend to use in assembling this
				model.
			container: Container or list of Containers. The container(s) to
				assemble in a model. If this is a list, then note that order is
				important, since layers without explicit 'input' connections
				will be connected to the previous layer.
		"""
		if isinstance(containers, Container):
			containers = [containers]
		self.root = ContainerGroup(containers)

		self.backend = backend

		self.network = None
		self.inputs = None
		self.outputs = None
		self.endpoints = None
		self.container_nodes = None

		self._parsed = False

		self.input_aliases = {}
		self.output_aliases = {}
		self.key_cache = {}

		self.provider = None
		self.additional_sources = {}

		self.compiled = None
		self.data = self.backend.create_data(self)

	###########################################################################
	def has_data_source(self, name):
		""" Checks if an auxiliary data source is present.
		"""
		return name in self.additional_sources

	###########################################################################
	def add_data_source(self, name, source):
		""" Registers an auxiliary data source.
		"""
		self.additional_sources[name] = source

	###########################################################################
	def get_data_sources(self):
		""" Returns an iterator over (key, value) tuples of auxiliary data
			sources.
		"""
		return self.additional_sources.items()

	###########################################################################
	def register_provider(self, provider):
		""" Let's the model know which data provider we are using.

			This allows the model to do shape inference for layers.
		"""
		self.provider = provider

	###########################################################################
	def supplement_provider(self, provider):
		for name, source in self.get_data_sources():
			provider.add_source(source, name=name)

	###########################################################################
	def get_inferred_shape(self, name):
		""" Tries to infer the shape of a container from the data.

			# Arguments

			name: str. The name of the data source.

			# Return value

			If a data provider has been registered (through a previous call to
			`register_provider()`, and if `name` is one of that provider's data
			sources, then the shape of that source is returned (as a tuple).
			Otherwise, None is returned.
		"""
		logger.debug('Trying to infer shape for input "%s"', name)
		if self.provider is None:
			logger.debug(
				'No provider has been registered to use for shape inference.')
			return None
		if self.provider.keys is None:
			logger.debug(
				'The provider does not have data keys associated with it.')
			return None

		try:
			index = self.provider.keys.index(name)
		except ValueError:
			logger.debug(
				'No such data source found for shape inference: %s', name)
			return None

		shape = self.provider.sources[index].shape()
		logger.debug('Inferred shape for input "%s": %s', name, shape)
		return shape

	###########################################################################
	def get_data_name_by_layer_name(self, keys, layer_name,
		key_cache=None, aliases=None):
		""" Finds the key of a data dictionary which is intended to provide
			data to a particular layer in the model.

			# Example

			```python
			# Say that you have data of this form:
			data = {'Aaron' : _, 'Barbara' : _, 'Charles' : _}
			# And the set of model aliases is:
			input_aliases = {'Apple' : 'A', 'Aaron' : 'A', 'Beth' : 'B', ...}
			# Then:
			assert get_data_name_by_layer_name(
				data.keys(),
				'A'
			) == 'Aaron'
			```
		"""
		key_cache = key_cache or self.key_cache
		aliases = aliases or (self.input_aliases, self.output_aliases)

		key = key_cache.get(layer_name)
		if key in keys:
			return key

		for alias_list in aliases:
			for key, name in alias_list.items():
				if key in keys and name == layer_name:
					key_cache[layer_name] = key
					return key

		return layer_name

	###########################################################################
	def get_layer_name_by_data_name(self, data_name):
		""" Finds the canonical name of the layer which uses a particular data
			source.
		"""
		if data_name in self.input_aliases:
			return self.input_aliases[data_name]
		if data_name in self.output_aliases:
			return self.output_aliases[data_name]
		logger.error('No such data name found: "%s". Something will probably '
			'break in a moment. This is probably a bug.', data_name)
		return data_name

	###########################################################################
	def get_backend(self):
		""" Returns the backend this model is using.
		"""
		return self.backend

	###########################################################################
	def save(self, filename):
		""" Saves the model weights to the given filename.

			# Arguments

			filename: str. The filename to write the weights to.

			# Notes

			The file format is backend-specific. There is no guarantee of
			compatability between different backends.
		"""
		logger.debug('Saving model weights to: %s', filename)
		self.backend.save(self, filename)

	###########################################################################
	def restore(self, filename):
		""" Load the model weights from the given filename.

			# Arguments

			filename: str. The filename to read the weights from.

			# Notes

			The file format is backend-specific. There is no guarantee of
			compatability between different backends.
		"""
		logger.debug('Loading model weights from: %s', filename)
		self.backend.restore(self, filename)

	###########################################################################
	def parse(self, engine):
		""" Parses the model.
		"""
		if not self._parsed:
			if engine is None:
				logger.info('Creating a dummy engine for parsing the model.')
				engine = PassthroughEngine()
			self.root.parse(engine)
			self._parsed = True

	###########################################################################
	def is_built(self):
		""" Returns True if this model has been built at some point.
		"""
		return self.network is not None

	###########################################################################
	def build(self):
		""" Builds the model.
		"""

		if not self._parsed:
			logger.warning('The model has not been parsed yet. We will try to '
				'parse it without context, but this may easily fail. Make '
				'sure Model.parse() is called before Model.build().')
			self.parse(None)

		logger.info('Enumerating the model containers.')

		# Construct the high-level network nodes.
		nodes = self.enumerate_nodes(self.root)

		logger.info('Assembling the model dependency graph.')
		input_nodes, output_nodes, network = self.assemble_graph(nodes)

		if logger.isEnabledFor(logging.DEBUG):
			queue = deque(input_nodes.values())
			while queue:
				node = queue.popleft()
				logger.debug('Assembled Node: %s', node.container.name)
				logger.debug('  Uses: %s', ', '
					.join([x.container.name for x in node.inputs]))
				logger.debug('  Used by: %s', ', '
					.join([x.container.name for x in node.outputs]))
				logger.debug('  Aliases: %s', ', '.join(node.names))
				queue.extend(node.outputs)

		logger.info('Connecting the model graph.')
		inputs, input_aliases, outputs, output_aliases = \
			self.build_graph(input_nodes, output_nodes, network)

		logger.info('Model inputs:  %s', ', '.join(node for node in inputs))
		logger.info('Model outputs: %s', ', '.join(node for node in outputs))

		self.inputs = inputs
		self.outputs = outputs
		self.network = network
		self.input_aliases = input_aliases
		self.output_aliases = output_aliases
		self.key_cache = {}

		self.compiled = None

	###########################################################################
	def build_graph(self, input_nodes, output_nodes, network):
		""" Builds and connects the model's underlying tensor operations.
		"""
		from ..utils.flatiter import flatten

		# 1-to-1 mapping: canonical name to compiled object.
		inputs = OrderedDict()
		# Many-to-one map: every name the input might have should map to the
		# key used in `inputs`.
		input_aliases = {}
		outputs = OrderedDict()
		output_aliases = {}

		for node in network.values():
			logger.debug('Building node: %s', node.container.name)
			logger.debug('  Aliases: %s', ', '.join(node.names))
			logger.debug('  Inputs:')
			for x in node.inputs:
				logger.debug('  - %s: %s', x.container.name, x.value)

			if node.inputs:

				# If you say that node N gets input from node M, you expect a
				# single tensor to be produced by M. After all, N is also going
				# to produce a single tensor. So you can always expect previous
				# nodes to produce single tensors.
				# If the node M's value is None, it hasn't modified the inputs
				# in any way. But if it only can produce a single output, then
				# not modifying the inputs implies that is also has only one
				# input.

				value = list(flatten([
					x.value for x in node.inputs
				]))
				meat = False

				for layer in node.container.build(self):
					if layer is None:
						continue
					meat = True
					try:
						value = self.backend.connect(
							inputs=value,
							target=layer,
							data=self.data
						)
					except:
						logger.error('Failed to connect node "%s" to its '
							'inputs: %s', node.container.name, ', '.join(
							x.container.name for x in node.inputs))
						logger.error('The input values are:')
						for i, v in zip(node.inputs, value):
							logger.error('%s = %s', i.container.name, v)
						raise

				if meat:
					node.value = value
				else:
					if len(value) != 1:
						raise ValueError('You cannot have an "empty" layer '
							'with more than one input. The "{}" layer has {} '
							'input layers.'.format(node.container.name,
							len(value)))
					node.value = value[0]
					node.names.extend(node.inputs[0].names)
					node.inputs[0].names = []

			else:
				value = None
				for layer in node.container.build(self):
					if layer is None:
						continue
					if value is None:
						value = layer
						# Register inputs
						inputs[node.container.name] = node #value
						for name in node.names:
							input_aliases[name] = node.container.name
					else:
						value = self.backend.connect(
							inputs=value,
							target=layer,
							data=self.data
						)
				# Comments:
				# - If your layer didn't produce anything, then `value` is
				#   None, and the `inputs` for this layer is set to None.
				# - If you decide to have your first layers produce None, then
				#   presumably you can figure out what to do with that value
				#   in upstream calls to `Backend.connect()`.
				# - If this layer produced nothing at all (not simply a None
				#   layer, but truly no layers at all), then `value` is still
				#   None

				node.value = value

			logger.debug('  Value: %s', node.value)

			# Register outputs
			if node.container.name in output_nodes:
				outputs[node.container.name] = node
				for name in node.names:
					output_aliases[name] = node.container.name

		return inputs, input_aliases, outputs, output_aliases

	###########################################################################
	def assemble_graph(self, nodes):
		""" Creates the dependency graph of containers in the model.
		"""

		name_map = {name : node.container
			for node in nodes for name in node.names}

		# Instantiate the network.
		# Map names to resolved nodes.
		network = OrderedDict()
		network.by_name = {}
		for node in nodes:
			container_node = ContainerNode(
				inputs=[],
				container=node.container,
				outputs=[],
				names=node.names,
				value=None
			)
			network[node.container.name] = container_node
			for name in node.names:
				network.by_name[name] = container_node

		# Populate the inputs.
		recent = deque()
		for node in nodes:
			# Get the new node.
			resolved = network[node.container.name]

			# Check if it has inputs.
			if node.inputs:
				if isinstance(node.container, Placeholder):
					raise ValueError('Input nodes cannot themselves have '
						'inputs. It looks like container "{}" tried to, '
						'though.'.format(node.container.name))
				# It has explicit inputs. Use them.
				for name in node.inputs:
					# This name is not necessarily the container's canonical
					# name. Let's find it.
					if name in name_map:
						# Find the container by a name.
						container = name_map[name]
						# Get the ContainerNode instance by canonical name.
						input_node = network[container.name]
						# Add the input.
						resolved.inputs.append(input_node)
					else:
						# Maybe this is supposed to be an input?
						raise ValueError('Container "{}" is trying to attach '
							'to input "{}", but we cannot find that input.'
							.format(node.container.name, name))
			else:
				# It does not have explicit inputs.
				if isinstance(node.container, Placeholder):
					# No worries, it's just a new input.
					recent = deque()
				elif recent:
					# There is something to connect to behind this node.
					resolved.inputs.append(recent[-1])
				else:
					# There is nothing there.
					# This should be an input layer.
					raise ValueError('Container "{}" looks like it is '
						'supposed to be an input layer. If that is true, '
						'then it ought to be marked as "{}".'
						.format(
							node.container.name,
							Placeholder.get_container_name()
						))

			recent.append(resolved)

		# Populate the outputs.
		for node in network.values():
			for input_node in node.inputs:
				input_node.outputs.append(node)

		used_as_output = set()

		input_names = set()
		output_names = set()
		for name, node in network.items():
			if not node.inputs:
				input_names.add(name)
			else:
				# Mark its inputs as not outputs.
				for input_node in node.inputs:
					used_as_output.add(input_node.container.name)

			if node.container.sink:
				output_names.add(name)

		for name in set(network.keys()) - used_as_output:
			output_names.add(name)

		input_nodes = {name : network[name] for name in input_names}
		output_nodes = {name : network[name] for name in output_names}

		# Last step: resolve the dependency graph.
		ordered = OrderedDict()
		# Pre-populate with input nodes.
		for k, v in network.items():
			if not v.inputs:
				ordered[k] = v
		for k in ordered:
			del network[k]
		# Now add the rest of the graph.
		while network:
			# Make the next pass through the nodes.
			changed = set()
			for k, v in network.items():
				for node in v.inputs:
					if node.container.name not in ordered:
						break
				else:
					ordered[k] = v
					changed.add(k)
			if not changed:
				raise ValueError('No change during dependency graph '
					'resolution. There is something wrong with the graph.')
			for k in changed:
				del network[k]
		ordered.by_name = network.by_name
		network = ordered

		return input_nodes, output_nodes, network

	###########################################################################
	def enumerate_nodes(self, root):
		""" Enumerates all of the layers in the model.
		"""
		if root.terminal():
			return [CollapsedContainer(
				inputs=root.inputs or [],
				container=root,
				names=[root.name] if root.name else []
			)]

		result = []
		for child in root.get_children(recursive=False):
			result.extend(self.enumerate_nodes(child))
		if result:
			if root.inputs:
				result[0].inputs = root.inputs
			if root.name:
				result[-1].names.append(root.name)
			if root.sink:
				result[-1].sink = True
		return result

	###########################################################################
	def get_shape_at_layer(self, name, assumptions=None):
		return self.get_shape_at_node(
			self.network.by_name[name],
			assumptions or {}
		)

	###########################################################################
	def get_shape_at_node(self, node, assumptions):
		for k, v in assumptions.items():
			if k in node.names:
				return v

		if node.inputs:
			return node.container.shape(
				input_shapes=[
					self.get_shape_at_node(input_node, assumptions)
					for input_node in node.inputs
				]
			)
		else:
			return node.container.shape(None)

### EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF



"""
Copyright 2017 Deepgram

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import os
import logging

import urllib.request

from . import TrainingHook, EvaluationHook
from .transcript import Transcript
from ...utils import prepare_multipart, prepare_json, UploadFile
from ...supplier import SpeechRecognitionSupplier

logger = logging.getLogger(__name__)

###############################################################################
class SlackHook(TrainingHook, EvaluationHook):
	""" Hook for posting to Slack.
	"""

	FILES_UPLOAD = 'https://slack.com/api/files.upload'

	###########################################################################
	@classmethod
	def get_name(cls):
		""" Returns the name of the hook.
		"""
		return 'slack'

	###########################################################################
	def __init__(self, channel, url=None, user=None, icon=None, title=None,
		token=None, extra_files=None, *args, **kwargs):
		""" Creates a new Slack hook.
		"""
		super().__init__(*args, **kwargs)

		if extra_files and isinstance(extra_files, str):
			extra_files = [extra_files]
		elif extra_files is None:
			extra_files = []

		self.username = user or 'kur'
		self.icon = icon or 'dragon'
		self.url = url
		self.channel = channel
		self.title = title
		self.token = token
		self.extra_files = extra_files

		if url is None and token is None:
			raise ValueError('Slack hook requires at least one of "url" or '
				'"token" is defined.')

		if extra_files and token is None:
			raise ValueError('A Slack "token" is required to upload files.')

	###########################################################################
	def upload_message(self, filename, text=None):
		""" Sends a message to Slack and uploads a file.
		"""
		logger.debug('Upload file: %s', filename)

		if not self.token:
			logger.warning('Skipping this upload -- "token" was not defined.')
			return

		if not os.path.isfile(filename):
			logger.warning('Skipping this upload -- file does not exist: %s',
				filename)
			return

		data = {
			'token' : self.token,
			'file' : UploadFile(filename),
			'filename' : filename,
			'channels' : self.channel
		}

		if text:
			data['initial_comment'] = text

		if self.title is not None:
			data['title'] = self.title

		data, header = prepare_multipart(data)
		self._submit(SlackHook.FILES_UPLOAD, data, header)

	###########################################################################
	def send_message(self, text, info=None):
		""" Sends a message to Slack.
		"""
		logger.debug('Sending Slack message.')

		if not self.url:
			logger.warning('Skipping this message -- "url" was not defined.')
			return

		if self.title is not None:
			text = '{}: {}'.format(self.title, text)

		if info:
			text = '{} More information: {}'.format(text,
				', '.join('{} = {}'.format(k, v) for k, v in info.items())
			)

		data = {
			'channel' : self.channel,
			'username' : self.username,
			'icon_emoji' : ':{}:'.format(self.icon),
			'text' : text
		}

		data, header = prepare_json(data)
		self._submit(self.url, data, header)

	###########################################################################
	def _submit(self, url, data, header):
		""" Submits a POST request to Slack.
		"""
		request = urllib.request.Request(
			url,
			data=data,
			headers=header,
			method='POST'
		)
		director = urllib.request.build_opener()

		try:
			response = director.open(request)
		except:									# pylint: disable=bare-except
			logger.exception('Failed to connect to Slack. Make sure the URL '
				'and channel are correct. If the channel was newly created, '
				'it might take a little time for Slack to catch up.')
		else:
			if response.code != 200:
				logger.error('Failed to post Slack notification. Make sure '
				'the URL and channel are correct. If the channel was newly '
				'created, it might take a little time for Slack to catch up.')

	###########################################################################
	def notify(self, status, log=None, info=None):
		""" Sends the Slack message.
		"""

		logger.debug('Slack hook received training message.')

		info = info or {}

		if status is TrainingHook.EPOCH_END:
			epoch = info.pop('epoch', None)
			total_epochs = info.pop('total_epochs', None)
			text = 'Finished epoch {} of {}.'.format(epoch, total_epochs)
		elif status is TrainingHook.TRAINING_END:
			text = 'Training has ended.'
		elif status is TrainingHook.TRAINING_START:
			text = 'Started training.'
		else:
			text = None

		if text:
			self.send_message(text, info)

		if status in (
			TrainingHook.EPOCH_END,
			TrainingHook.VALIDATION_END
		):
			self.upload_extra_files()

	###########################################################################
	def upload_extra_files(self):
		""" Uploads `extra_files` to Slack.
		"""
		if self.token:
			for filename in self.extra_files:
				if os.path.isfile(filename):
					self.upload_message(filename)
				else:
					logger.debug('Skipping file upload -- file does not '
						'exist: %s', filename)
		else:
			logger.debug('"token" is not defined. Skipping uploads.')

	###########################################################################
	def apply(self, current, original, model=None):
		""" Sends a Slack message in response to an evaluation hook.
		"""

		logger.debug('Slack hook received non-training message.')

		data, truth = current

		upload = False
		if self.token is not None:
			if isinstance(data, Transcript) \
					and original[1] is not None \
					and 'audio_source' in original[1]:
				path = SpeechRecognitionSupplier.find_audio_path(
					original[1]['audio_source'][0]
				)
				if path is not None:
					upload = True

		text = 'Truth = "{}", Prediction = "{}"'.format(truth, data),
		if upload:
			self.upload_message(path, text)
		elif self.url is not None:
			self.send_message(text)
		else:
			logger.warning('Failed to post to Slack. The "slack" hook was '
				'given enough information for uploading only, and not enough '
				'for message posting. However, the data you are working with '
				'does not provide enough information for file uploading.')

		self.upload_extra_files()

		return current

### EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF



"""
Copyright 2017 Deepgram

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from . import Layer, ParsingError

###############################################################################
class Merge(Layer):					# pylint: disable=too-few-public-methods
	""" A container for merging inputs from multiple input layers.
	"""

	MERGE_MODES = ('multiply', 'add', 'concat', 'average')
	DEFAULT_MERGE_MODE = 'concat'

	DEFAULT_AXIS = -1

	###########################################################################
	def __init__(self, *args, **kwargs):
		""" Create a new merge container.
		"""
		super().__init__(*args, **kwargs)
		self.mode = None
		self.axis = None

	###########################################################################
	def _parse(self, engine):
		""" Parse the child containers
		"""
		if isinstance(self.args, dict):
			self.mode = engine.evaluate(self.args.get('mode'), recursive=True)
			self.axis = engine.evaluate(self.args.get('axis'), recursive=True)
		else:
			self.mode = self.args
			self.axis = Merge.DEFAULT_AXIS

		if self.mode is None:
			self.mode = Merge.DEFAULT_MERGE_MODE
		if not isinstance(self.mode, str):
			raise ParsingError('Wrong type for "mode" argument in '
				'recurrent layer. Expected one of: {}. Received: {}'
					.format(', '.join(Merge.MERGE_MODES), self.mode)
				)
		self.mode = self.mode.lower()
		if self.mode not in Merge.MERGE_MODES:
			raise ParsingError('Bad value for "mode" argument in '
				'recurrent layer. Expected one of: {}. Received: {}'
					.format(', '.join(Merge.MERGE_MODES), self.mode)
				)

		if not isinstance(self.axis, int):
			raise ParsingError('Axis must be an integer. Received: {}'
				.format(self.axis))

	###########################################################################
	def _build(self, model):
		""" Instantiate the container.
		"""
		backend = model.get_backend()
		if backend.get_name() == 'keras':

			import keras.layers as L			# pylint: disable=import-error

			def merge(inputs):
				""" Merges inputs.
				"""
				# Keras "merge" requires more than one layer.
				if len(inputs) == 1:
					return inputs[0]

				if backend.keras_version() == 1:
					return L.merge(
						inputs,
						mode={
							'multiply' : 'mul',
							'add' : 'sum',
							'concat' : 'concat',
							'average' : 'ave'
						}.get(self.mode),
						name=self.name,
						**{
							'concat' : {'concat_axis' : self.axis}
						}.get(self.mode, {})
					)
				else:
					func = {
						'multiply' : L.multiply,
						'add' : L.add,
						'concat' : L.concatenate,
						'average' : L.average
					}.get(self.mode)
					return func(
						inputs,
						axis=self.axis,
						name=self.name
					)

			yield merge

		elif backend.get_name() == 'pytorch':

			import torch						# pylint: disable=import-error

			def connect(inputs):
				""" Connects the layer.
				"""
				axis = self.axis
				if axis < 0:
					axis += len(inputs[0]['shape'])
				axis += 1
				return {
					'shape' : self.shape([x['shape'] for x in inputs]),
					'layer' : model.data.add_operation(
						lambda *x: torch.cat(x, axis)
					)(*[x['layer'] for x in inputs])
				}

			yield connect

		else:
			raise ValueError(
				'Unknown or unsupported backend: {}'.format(backend)
			)

	###########################################################################
	def shape(self, input_shapes):
		""" Returns the output shape of this layer for a given input shape.
		"""
		if self.mode == 'concat':
			result = None
			for input_shape in input_shapes:
				if result is None:
					result = list(input_shape)
				else:
					result[self.axis] += input_shape[self.axis]
			return tuple(result)

		return input_shapes[0]

### EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF



"""
Copyright 2016 Deepgram

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import json
import re
import os
import logging
import random

import numpy

from ..sources import DerivedSource, VanillaSource, ChunkSource
from . import Supplier
from ..utils import package
from ..utils import count_lines
from ..utils import get_audio_features
from ..utils import Normalize

logger = logging.getLogger(__name__)

###############################################################################
def loop_copy(src, dest):
	""" Copies a source array into a destination array, looping
		through the source array until the entire destination array
		is written.

		# Arguments

		src: numpy array. The source array.
		dest: numpy array. The destination array.

		# Return value

		The destination array.

		# Notes

		If the source array is longer than the destination array, then
		only the first N entries of the source array are copied into the
		destination, where N is the length of the destination array.

		If the source array is shorter than, or of equal length with, the
		destination array, then it will be copied into the destination array
		and repeated, back-to-back, until the destination is filled up (almost
		tiling the destination, if you will).
	"""
	if len(src) > len(dest):
		dest[:] = src[:len(dest)]
	else:
		i = 0
		while i < len(dest):
			entries_to_copy = min(len(dest) - i, len(src))
			dest[i:i+entries_to_copy] = src[:entries_to_copy]
			i += len(src)

	return dest

###############################################################################
class UtteranceLength(DerivedSource):
	""" Data source for audio lengths.
	"""
	def __init__(self, source):
		super().__init__()
		self.source = source
	def derive(self, inputs):
		utterances, = inputs
		return numpy.array([[len(x)] for x in utterances], dtype='int32')
	def shape(self):
		return (1, )
	def requires(self):
		return (self.source, )

###############################################################################
class Utterance(DerivedSource):
	""" Data source for model-ready audio samples. Unlike `RawUtterance`, this
		ensures that all data products are rectangular tensors (rather than
		ragged arrays).
	"""
	def __init__(self, source, raw, fill=None):
		super().__init__()
		self.source = source
		self.raw = raw
		self.fill = fill or 'zero'
		assert isinstance(self.fill, str) and self.fill in ('zero', 'loop')
		logger.debug('Utterance source is using fill mode "%s".', self.fill)
	def derive(self, inputs):
		utterances, = inputs
		max_len = max(len(x) for x in utterances)

		if self.fill == 'zero':
			output = numpy.zeros(
				shape=(len(utterances), max_len, self.raw.features)
			)
			for i, row in enumerate(utterances):
				output[i][:len(row)] = row
		elif self.fill == 'loop':
			output = numpy.empty(
				shape=(len(utterances), max_len, self.raw.features)
			)
			for i, row in enumerate(utterances):
				loop_copy(row, output[i])
		else:
			raise ValueError('Unhandled fill type: "{}". This is a bug.'
				.format(self.fill))

		return output
	def shape(self):
		return (None, self.raw.features)
	def requires(self):
		return (self.source, )

###############################################################################
class RawUtterance(ChunkSource):
	""" Data source for audio samples
	"""

	DEFAULT_NORMALIZATION_DEPTH = 100

	###########################################################################
	@classmethod
	def default_chunk_size(cls):
		""" Returns the default chunk size for this source.
		"""
		return ChunkSource.USE_BATCH_SIZE

	###########################################################################
	def __init__(self, audio_paths, feature_type=None,
		normalization=None, max_frequency=None, *args, **kwargs):
		""" Creates a new raw utterance source.
		"""

		super().__init__(*args, **kwargs)
		self.audio_paths = audio_paths
		self.indices = numpy.arange(len(self))
		self.feature_type = feature_type
		self.features = None
		self.max_frequency = max_frequency

		self._init_normalizer(normalization)

	###########################################################################
	def _init_normalizer(self, params):

		# Parse the specification.
		if isinstance(params, str):
			params = {'path' : params}
		elif params is None:
			params = {'path' : None}
		elif not isinstance(params, dict):
			raise ValueError('Unknown normalization value: {}'.format(params))

		# Merge in the defaults.
		defaults = {
			'path' : None,
			'center' : True,
			'scale' : True,
			'rotate' : True,
			'depth' : RawUtterance.DEFAULT_NORMALIZATION_DEPTH
		}
		defaults.update(params)
		params = defaults

		# Create the normalizer.
		norm = Normalize(
			center=params['center'],
			scale=params['scale'],
			rotate=params['rotate']
		)

		path = params['path']
		if path is None:
			logger.info('No normalization data available. We will use '
				'on-the-fly (non-persistent) normalization. In the future, '
				'you probably want to give a filename to the "normalization" '
				'key in the speech recognition supplier.')
			self.train_normalizer(norm, depth=params['depth'])
		else:
			path = os.path.expanduser(os.path.expandvars(path))
			if os.path.exists(path):
				if not os.path.isfile(path):
					raise ValueError('Normalization data must be a regular '
						'file. This is not: {}'.format(path))
				logger.info('Restoring normalization statistics: %s', path)
				norm.restore(path)
				self.features = norm.get_dimensionality()
			else:
				logger.info('Training new normalization statistics: %s', path)
				self.train_normalizer(norm, depth=params['depth'])
				norm.save(path)

		# Register the normalizer
		self.norm = norm

	###########################################################################
	def load_audio(self, paths):
		""" Loads unnormalized audio data.
		"""
		# Resolve and load each path.
		result = [None] * len(paths)
		for i, partial_path in enumerate(paths):
			path = SpeechRecognitionSupplier.find_audio_path(partial_path)
			if path is None:
				logger.error('Could not find audio file that---ignoring '
					'extension---begins with: %s', partial_path)
			else:
				result[i] = get_audio_features(
					path,
					feature_type=self.feature_type,
					high_freq=self.max_frequency,
					on_error='suppress'
				)
				if result[i] is None:
					logger.error('Failed to load audio file at path: %s', path)

		# Clean up bad audio
		if any(x is None for x in result):
			logger.warning('Recovering from a bad audio uttereance.')
			good = None
			for candidate in result:
				if candidate is not None:
					good = candidate
					break
			if good is None:
				raise ValueError(
					'Cannot tolerate an entire batch of bad audio.')
			for i, x in enumerate(result):
				if x is None:
					result[i] = good

		return result

	###########################################################################
	def train_normalizer(self, norm, depth):
		""" Trains the normalizer on the data.

			# Arguments

			norm: Normalize instance. The normalization transform to train.
		"""
		logger.info('Training normalization transform.')
		num_entries = min(depth, len(self.audio_paths))
		paths = random.sample(self.audio_paths, num_entries)
		data = self.load_audio(paths)
		self.features = data[0].shape[-1]
		norm.learn(data)
		logger.debug('Finished training normalization transform.')

	###########################################################################
	def __iter__(self):
		""" Return an iterator to the data.
		"""
		start = 0
		num_entries = len(self)
		while start < num_entries:
			end = min(num_entries, start + self.chunk_size)

			paths = [self.audio_paths[i] for i in self.indices[start:end]]
			batch = self.load_audio(paths)
			batch = [self.norm.apply(data) for data in batch]
			yield batch
			start = end

	###########################################################################
	def __len__(self):
		""" Returns the total number of entries that this source can return, if
			known.
		"""
		return len(self.audio_paths)

	###########################################################################
	def shape(self):
		""" Return the shape of the tensor (excluding batch size) returned by
			this data source.
		"""
		return (None, self.features)

	###########################################################################
	def can_shuffle(self):
		""" This source can be shuffled.
		"""
		return True

	###########################################################################
	def shuffle(self, indices):
		""" Applies a permutation to the data.
		"""
		if len(indices) > len(self):
			raise ValueError('Shuffleable was asked to apply permutation, but '
				'the permutation is longer than the length of the data set.')
		self.indices[:len(indices)] = self.indices[:len(indices)][indices]

###############################################################################
class TranscriptLength(DerivedSource):
	""" Data source for computing transcript lengths.
	"""
	def __init__(self, source):
		super().__init__()
		self.source = source
	def derive(self, inputs):
		transcript, = inputs
		return numpy.array([[len(x)] for x in transcript], dtype='int32')
	def shape(self):
		return (1, )
	def requires(self):
		return (self.source, )

###############################################################################
class Transcript(DerivedSource):
	""" Data source for neat (non-ragged) transcript arrays.
	"""
	def __init__(self, source):
		super().__init__()
		self.source = source
	def derive(self, inputs):
		transcript, = inputs
		max_len = max(len(x) for x in transcript)
		output = numpy.zeros(shape=(len(transcript), max_len), dtype='int32')
		for i, row in enumerate(transcript):
			output[i][:len(row)] = row
		return output
	def shape(self):
		return (None, )
	def requires(self):
		return (self.source, )

###############################################################################
class RawTranscript(ChunkSource):
	""" Data source for variable-length transcripts.
	"""

	###########################################################################
	def __init__(self, transcripts, vocab=None, *args,
		**kwargs):
		""" Creates a new raw transcript source.
		"""
		super().__init__(*args, **kwargs)
		self.transcripts = transcripts
		self.indices = numpy.arange(len(self))
		self.vocab = self.make_vocab(vocab)

	###########################################################################
	def make_vocab(self, vocab):
		""" Loads or infers a vocabulary.
		"""
		if vocab is None:
			logger.info('Inferring vocabulary from data set.')
			data = set(x for transcript in self.transcripts \
				for x in transcript.lower())
			data = sorted(data)

		elif isinstance(vocab, str):
			logger.info('Load vocabulary from a JSON file: %s', vocab)
			with open(vocab) as fh:
				json_data = json.loads(fh.read())
			try:
				data = [x.lower() for x in json_data]
			except:
				logger.exception('Expected the JSON to contain a single list '
					'of strings. Instead, we got: %s', json_data)
				raise

		elif isinstance(vocab, (tuple, list)):
			logger.info('Using a hard-coded vocabulary.')
			try:
				data = [x.lower() for x in vocab]
			except:
				logger.exception('Expected the vocabulary to be a list of '
					'strings. Instead, we got: %s', vocab)
				raise

		else:
			raise ValueError('Unknown vocabulary format: {}'.format(vocab))

		if len(set(data)) != len(data):
			raise ValueError('The vocabulary must contain unique entries, but '
				'we found duplicates. Make sure that all entries are unique, '
				'ignoring capitalization. That means you should not have both '
				'"x" and "X" in your vocabulary. For reference, this is the '
				'vocabulary we ended up with: {}'.format(data))

		logger.info('Loaded a %d-word vocabulary.', len(data))
		return {x : i for i, x in enumerate(data)}

	###########################################################################
	def word_to_integer(self, data):
		""" Maps a character transcript to its integer representation.
		"""
		result = [None]*len(data)
		for i, row in enumerate(data):
			result[i] = [self.vocab.get(word.lower()) for word in row]
			result[i] = [x for x in result[i] if x is not None]
		return result

	###########################################################################
	def __iter__(self):
		""" Return an iterator to the data.
		"""
		start = 0
		num_entries = len(self)
		while start < num_entries:
			end = min(num_entries, start + self.chunk_size)

			batch = [self.transcripts[i] for i in self.indices[start:end]]
			batch = self.word_to_integer(batch)
			yield batch
			start = end

	###########################################################################
	def __len__(self):
		""" Returns the total number of entries that this source can return, if
			known.
		"""
		return len(self.transcripts)

	###########################################################################
	def shape(self):
		""" Return the shape of the tensor (excluding batch size) returned by
			this data source.
		"""
		return (None, )

	###########################################################################
	def can_shuffle(self):
		""" This source can be shuffled.
		"""
		return True

	###########################################################################
	def shuffle(self, indices):
		""" Applies a permutation to the data.
		"""
		if len(indices) > len(self):
			raise ValueError('Shuffleable was asked to apply permutation, but '
				'the permutation is longer than the length of the data set.')
		self.indices[:len(indices)] = self.indices[:len(indices)][indices]

###############################################################################
class SpeechRecognitionSupplier(Supplier):
	""" A supplier which handles parsing of audio + transcript data sets for
		speech recognition purposes.
	"""

	DEFAULT_UNPACK = True
	DEFAULT_TYPE = 'spec'
	SUPPORTED_TYPES = ('wav', 'mp3', 'flac')

	###########################################################################
	@staticmethod
	def find_audio_path(partial_path):
		""" Resolves the audio file extension.
		"""
		for ext in SpeechRecognitionSupplier.SUPPORTED_TYPES:
			candidate = '{}.{}'.format(partial_path, ext)
			if os.path.isfile(candidate):
				return candidate
		return None

	###########################################################################
	@classmethod
	def get_name(cls):
		""" Returns the name of the supplier.
		"""
		return 'speech_recognition'

	###########################################################################
	def __init__(self, url=None, path=None, checksum=None, unpack=None, 
		type=None, normalization=None, min_duration=None, max_duration=None,
		max_frequency=None, vocab=None, samples=None, fill=None, *args,
		**kwargs):
		""" Creates a new speech recognition supplier.

			# Arguments
		"""
		super().__init__(*args, **kwargs)

		if unpack is None:
			unpack = SpeechRecognitionSupplier.DEFAULT_UNPACK
		self.load_data(url=url, path=path, checksum=checksum, unpack=unpack,
			min_duration=min_duration, max_duration=max_duration)
		self.downselect(samples)

		logger.debug('Creating sources.')
		utterance_raw = RawUtterance(
			self.data['audio'],
			feature_type=type or SpeechRecognitionSupplier.DEFAULT_TYPE,
			normalization=normalization,
			max_frequency=max_frequency
		)
		self.sources = {
			'transcript_raw' : RawTranscript(
				self.data['transcript'],
				vocab=vocab
			),
			'transcript_length' : TranscriptLength('transcript_raw'),
			'transcript' : Transcript('transcript_raw'),
			'utterance_raw' : utterance_raw,
			'utterance_length' : UtteranceLength('utterance_raw'),
			'utterance' : Utterance('utterance_raw', utterance_raw, fill=fill),
			'duration' : VanillaSource(numpy.array(self.data['duration'])),
			'audio_source' : VanillaSource(numpy.array(self.data['audio']))
		}

	###########################################################################
	def downselect(self, samples):
		""" Selects a subset of the data.

			# Arguments

			samples: None, int, or str. If None, uses all samples. If an
				integer, uses the first `samples` entries. If a string, follows
				the `Sample Specification` below.

			# Sample Specification

			Forms:			Meaning:

			10				Use 10 samples: 0 through 10.
			10-				Use all but 10 samples: use the 10th through
							the remainder.
			10%				Use the first 10% of samples.
			10-20			Use 10 samples: 10 through 19.
			10-20%			Use 10% of samples, from 10% through 20%.
							When combined with a random seed, this lets you
							split a dataset on the fly.
			10-%			Use all but 10% of samples: from 10% through
							the remainder.

			# Notes:

			- For percentage ranges of the form "X-%" or "X-Y%", Kur will
			  compute the percentage and then add one to the start value.
			  This makes it easier for you to use a random seed and then
			  make dataset splits like 10%, 10-20%, 20% without worrying
			  about having disjoint datasets.

			- As evidenced by the examples above, Kur follows Python in using
			  ranges that exclude the upper bound (e.g., "10" means "0 .. 9"
			  and "10-20" means "10..19").
		"""
		if samples is None:
			logger.debug('Using all available data.')
			return
		elif isinstance(samples, int):
			if samples < 1:
				raise ValueError('"samples" cannot be less than 1.')
			if samples >= self.metadata['entries']:
				return
			logger.debug('Using only %d / %d samples of the available data.',
				samples, self.metadata['entries'])
			start = 0
			end = samples
		elif isinstance(samples, str):
			regex = re.compile(
				r'(?P<start>[0-9]+(?:\.[0-9]*)?)'
				r'(?:(?P<range>-)'
					r'(?P<end>[0-9]+(?:\.[0-9]*)?)?'
				r')?'
				r'(?P<unit>%)?'
			)
			match = regex.match(samples)
			if not match:
				raise ValueError('Failed to parse the "samples" '
					'specification: {}'.format(samples))

			result = match.groupdict()
			start = float(result['start'])
			if result['range']:
				if result['end']:
					end = float(result['end'])
				elif result['unit']:
					end = 100
				else:
					end = self.metadata['entries']
			else:
				end = start
				start = 0

			if result['unit']:
				start = int(self.metadata['entries'] * (start / 100))
				end = int(self.metadata['entries'] * (end / 100))
			else:
				start = int(start)
				end = int(end)

			start = min(max(0, start), self.metadata['entries'])
			end = min(max(0, end), self.metadata['entries'])

			if start == 0 and end == self.metadata['entries']:
				return

			if start >= end:
				raise ValueError('No samples pass this "samples" cut: [{}, {})'
					.format(start, end))
		else:
			raise TypeError('Invalid/unexpected type for "samples": {}'
				.format(samples))

		# Create the seeded random number generator.
		gen = numpy.random.RandomState(
			seed=self.kurfile.get_seed() if self.kurfile else None
		)

		# Produce a mask (True = keep, False = discard)
		mask = numpy.zeros(self.metadata['entries'], dtype=bool)
		indices = gen.permutation(self.metadata['entries'])[start:end]
		mask[indices] = True

		# Downselect
		for k in self.data:
			self.data[k] = [x for i, x in enumerate(self.data[k]) if mask[i]]

		self.metadata['entries'] = int(end - start)

	###########################################################################
	def load_data(self, url=None, path=None, checksum=None, unpack=None,
		min_duration=None, max_duration=None):
		""" Loads the data for this supplier.
		"""
		local_path, is_packed = package.install(
			url=url,
			path=path,
			checksum=checksum
		)

		manifest = None
		if is_packed and unpack:
			logger.debug('Unpacking input data: %s', local_path)
			manifest = package.unpack(local_path, recursive=True)
			is_packed = False
		elif is_packed and not unpack:
			logger.debug('Using packed input data.')
			raise NotImplementedError
		elif not is_packed and unpack:
			logger.debug('Using already unpacked input data.')
		elif not is_packed and not unpack:
			logger.debug('Ignore "unpack" for input data, since it is already '
				'unpacked.')
		else:
			logger.error('Unhandled data package requirements. This is a bug.')

		self.metadata, self.data = self.get_metadata(
			manifest=manifest,
			root=local_path,
			min_duration=min_duration,
			max_duration=max_duration
		)

	###########################################################################
	def get_metadata(self, manifest=None, root=None, 
		min_duration=None, max_duration=None):
		""" Scans the package for a metadata file, makes sure everything is in
			order, and returns some information about the data set.
		"""
		logger.debug('Looking for metadata file.')
		metadata_file = None

		def look_in_list(filenames):
			""" Searches a list of files for a JSONL file.
			"""
			for filename in filenames:
				parts = os.path.splitext(filename)
				if parts[1].lower() == '.jsonl' and \
					not os.path.basename(filename).startswith('.'):
					return filename
			return None

		if manifest is None:
			if root is None:
				raise ValueError('No root provided and no manifest provided. '
					'This is a bug.')
			if not os.path.isdir(root):
				raise ValueError('Root is not a directory. This is a bug.')
			for dirpath, _, filenames in os.walk(root):
				metadata_file = look_in_list(filenames)
				if metadata_file is not None:
					metadata_file = os.path.join(dirpath, metadata_file)
					break
		else:
			metadata_file = look_in_list(manifest)

		if metadata_file is None:
			raise ValueError('Failed to find a JSONL metadata file.')

		source = os.path.join(
			os.path.dirname(metadata_file),
			'audio'
		)

		logger.debug('Found metadata file: %s', metadata_file)
		logger.debug('Inferred source path: %s', source)

		logger.debug('Scanning metadata file.')
		lines = count_lines(metadata_file)
		logger.debug('Entries counted: %d', lines)

		logger.debug('Loading metadata.')
		data = {
			'audio' : [None]*lines,
			'transcript' : [None]*lines,
			'duration' : [None]*lines
		}

		entries = 0
		with open(metadata_file, 'r') as fh:
			for line_number, line in enumerate(fh, 1):
				try:
					entry = json.loads(line)
				except json.decoder.JSONDecodeError:
					logger.warning('Failed to parse valid JSON on line %d of '
						'file %s', line_number, metadata_file)
					continue
				if any(k not in entry for k in ('text', 'duration_s', 'uuid')):
					logger.warning('Line %d is missing one of its required '
						'keys in metadata file %s', line_number, metadata_file)
					continue

				duration = entry['duration_s']
				if min_duration and duration < min_duration:
					continue
				if max_duration and duration > max_duration:
					continue

				data['duration'][entries] = entry['duration_s']
				data['transcript'][entries] = entry['text']
				data['audio'][entries] = os.path.join(source, entry['uuid'])

				entries += 1

		logger.debug('Entries kept: %d', entries)
		for k in data:
			data[k] = data[k][:entries]

		metadata = {
			'entries' : entries,
			'filename' : metadata_file,
			'source' : source
		}

		return metadata, data

	###########################################################################
	def get_sources(self, sources=None):
		""" Returns all sources from this provider.
		"""

		if sources is None:
			sources = list(self.sources.keys())
		elif not isinstance(sources, (list, tuple)):
			sources = [sources]

		for source in sources:
			if source not in self.sources:
				raise KeyError(
					'Invalid data key: {}. Valid keys are: {}'.format(
						source, ', '.join(str(k) for k in self.sources.keys())
				))

		return {k : self.sources[k] for k in sources}

### EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF



"""
Copyright 2016 Deepgram

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from ..utils import get_subclasses

###############################################################################
def keras_wrap(model, target, output, loss):
	""" Convenience function for wrapping a Keras loss function.
	"""
	# pylint: disable=import-error
	import keras.objectives as O
	import keras.backend as K
	# pylint: enable=import-error
	if isinstance(loss, str):
		loss = O.get(loss)
	shape = model.outputs[target].value._keras_shape # pylint: disable=protected-access
	ins = [
		(target, K.placeholder(
			ndim=len(shape),
			dtype=K.dtype(model.outputs[target].value),
			name=target
		))
	]
	out = loss(ins[0][1], output)
	return ins, out

###############################################################################
class Loss:
	""" Base class for all loss functions (also called objective functions).
	"""

	###########################################################################
	@classmethod
	def get_name(cls):
		""" Returns the name of the loss function.

			# Return value

			A lower-case string unique to this loss function.
		"""
		return cls.__name__.lower()

	###########################################################################
	@staticmethod
	def get_all_losses():
		""" Returns an iterator to the names of all loss functions.
		"""
		for cls in get_subclasses(Loss):
			yield cls

	###########################################################################
	@staticmethod
	def get_loss_by_name(name):
		""" Finds a loss function class with the given name.
		"""
		name = name.lower()
		for cls in Loss.get_all_losses():
			if cls.get_name() == name:
				return cls
		raise ValueError('No such loss function with name "{}"'.format(name))

	###########################################################################
	def __init__(self, weight=None):
		""" Creates a new loss function.

			# Arguments

			weight: float. A relative weight to apply to this loss function.
				This is only meaningful in models which have multiple loss
				functions.
		"""
		if weight is not None:
			raise ValueError('Loss function weights have not been implemented '
				'yet.')
		self.weight = 1.0 if weight is None else weight

	###########################################################################
	def get_weight(self):
		""" Returns the loss function weight.
		"""
		return self.weight

	###########################################################################
	def get_loss(self, model, target, output):
		""" Returns the loss tensor for this output.

			# Arguments

			model: Model instance.
			target: str. The name of the output layer to apply the loss
				function to.
			output: tensor (implemented-specific). The symbolic tensor for this
				output layer.

			# Return value

			A tuple of the form:

			```python
			(
				# Input tensors
				[
					(input_name, placeholder),
					(input_name, placeholder),
					...
				],

				# Output value
				loss_value
			)
			```

			The derived class is required to return all required input
			placeholder, including placeholders for the target model outputs.
		"""
		raise NotImplementedError

### EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF



"""
Copyright 2016 Deepgram

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import logging
import warnings
from . import Layer, ParsingError

logger = logging.getLogger(__name__)

###############################################################################
class Placeholder(Layer):				# pylint: disable=too-few-public-methods
	""" A shape placeholder which serves as a size hint, typically used as the
		first layer for each input to the model.

		Placeholders are also convenient places to name an input stream.
	"""

	###########################################################################
	@classmethod
	def get_container_name(cls):
		""" Returns the name of the container class.
		"""
		return 'input'

	###########################################################################
	def __init__(self, *args, **kwargs):
		""" Creates a new placeholder.
		"""
		super().__init__(*args, **kwargs)
		self._shape = None
		self.type = None

	###########################################################################
	def _parse_pre(self, engine):
		""" Pre-parsing hook.
		"""
		super()._parse_pre(engine)

		container_name = self.get_container_name()
		if container_name in self.data:
			data = engine.evaluate(self.data[container_name])
			if isinstance(data, str):
				if 'name' in self.data:
					name = engine.evaluate(self.data['name'])
					logger.warning('Conflicting naming schemes for '
						'placeholder: "%s" and "%s". Using: "%s".',
						data, name, name)
				else:
					logger.debug('Using short-hand name for placeholder: %s',
						data)
					self.name = data
		else:
			warnings.warn('Parsing oddity in placeholder: {}'.format(
				self.data
			))

	###########################################################################
	def set_shape(self, shape):
		""" Sets a shape.
		"""
		if self._shape is not None:
			logger.warning('Modifying the shape of Placeholder "%s".',
				self.name)

		if not isinstance(shape, (list, tuple)):
			shape = (shape, )
		shape = tuple(x if x != 'None' else None for x in shape)
		for x in shape:
			if not isinstance(x, (int, type(None))):
				raise ParsingError(
					'All entries in "shape" must be integers, or in special '
					'cases None. Shape is: {}'.format(shape)
				)
		self._shape = shape

	###########################################################################
	def _parse(self, engine):
		""" Parse the placeholder.
		"""
		if 'shape' not in self.args:
			logger.debug('Placeholder "%s" has a deferred shape.', self.name)
		else:
			self.set_shape(engine.evaluate(self.args['shape'], recursive=True))

		if 'type' in self.args:
			self.type = engine.evaluate(self.args['type'], recursive=True)

	###########################################################################
	def _infer_shape(self, model):
		inferred_shape = model.get_inferred_shape(self.name)
		if inferred_shape is None:
			if self._shape is None:
				raise ParsingError(
					'Placeholder "{}" requires a shape.'.format(self.name))
		else:
			if self._shape is None:
				self._shape = inferred_shape
				logger.debug('Inferred shape: %s', self._shape)
			else:
				if len(self._shape) != len(inferred_shape):
					raise ValueError('Placeholder "{}" does not have the '
						'same dimensionality as the data source it is '
						'connected to. Placeholder: {}, data source: {}.'
						.format(self.name, self._shape, inferred_shape))

				merged_shape = ()
				for user_shape, data_shape in \
					zip(self._shape, inferred_shape):
					if user_shape is None or data_shape is None:
						if user_shape is None:
							merged_shape += (data_shape, )
						else:
							merged_shape += (user_shape, )
					elif user_shape != data_shape:
						logger.warning('Placeholder "%s" specified a '
							'along a dimension that disagrees with the '
							'data source. We will defer to the data '
							'source, but this may cause an unexpected '
							'model to be built. Placeholder: %s, data: '
							'%s.', self.name, self._shape, inferred_shape)
						merged_shape += (data_shape, )
					else:
						merged_shape += (data_shape, )

				if merged_shape != self._shape:
					logger.debug('Using inferred shape for data %s: %s '
						'instead of %s', self.name, merged_shape,
						self._shape)
					self._shape = merged_shape

		if self._shape == ():
			logger.debug('Promoting an empty shape to a scalar.')
			self._shape = (1,)

	###########################################################################
	def _build(self, model):
		""" Create the backend-specific placeholder.
		"""
		backend = model.get_backend()
		if backend.get_name() == 'keras':

			import keras.backend as K			# pylint: disable=import-error
			if self.type is None:
				dtype = K.floatx()
			else:
				dtype = self.type
			logger.debug('Creating placeholder for "%s" with data type "%s".',
				self.name, dtype)

			self._infer_shape(model)

			import keras.layers as L			# pylint: disable=import-error
			yield L.Input(
				shape=self._shape,
				name=self.name,
				dtype=dtype
			)

		elif backend.get_name() == 'pytorch':

			self._infer_shape(model)
			yield {
				'shape' : self._shape,
				'layer' : model.data.placeholder(self.name)
			}

		else:
			raise ValueError(
				'Unknown or unsupported backend: {}'.format(backend))

	###########################################################################
	def shape(self, input_shapes=None):
		""" Returns the output shape of this layer for a given input shape.
		"""
		if input_shapes is not None:
			raise ValueError('Input placeholders do not take inputs.')
		return self._shape

### EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF



"""
Copyright 2016 Deepgram

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import os

###############################################################################
class EnvironmentalVariable:		# pylint: disable=too-few-public-methods
	""" Context management for modifying environmental variables.
	"""

	###########################################################################
	def __init__(self, **kwargs):
		""" Creates a new EnvironmentalVariable manager.

			# Arguments

			**kwargs: key/value pairs to use in setting environmental variables
				inside the context.
		"""
		self.vars = kwargs
		self.old_vars = None

	###########################################################################
	def __enter__(self):
		""" Enters the context, setting the temporary environmental variables.
		"""
		self.push()

	###########################################################################
	def push(self):
		""" Sets the environmental variables. Idempotent.
		"""
		if self.old_vars is not None:
			return

		self.old_vars = {}
		for k, v in self.vars.items():
			self.old_vars[k] = os.environ.get(k)
			if v is None:
				if k in os.environ:
					del os.environ[k]
			else:
				os.environ[k] = v

	###########################################################################
	def __exit__(self, exc_type, exc_value, traceback):
		""" Exists the context, restoring the original environmental variables.
		"""
		self.pop()

	###########################################################################
	def pop(self):
		""" Reets the environmental variables. Idempotent.
		"""
		if self.old_vars is None:
			return

		for k, v in self.old_vars.items():
			if v is None:
				if k in os.environ:
					del os.environ[k]
			else:
				os.environ[k] = v

		self.old_vars = None

### EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF



"""
Copyright 2016 Deepgram

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

# length of the context sequence
seq_len = 30

lowercase_letters = [
    chr(97 + i) for i in range(26)
]
symbols = [' ', '"', '\'', '.']

# these are the characters we allow in our data
vocab = lowercase_letters + symbols

# we can convert between our vocab and integers
char_to_int = dict(
    (c, i) for i, c in enumerate(vocab)
)
int_to_char = dict(enumerate(vocab))

n_vocab = len(vocab)



"""
Copyright 2017 Deepgram

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import numpy

from . import ChunkSource

###############################################################################
class StackSource(ChunkSource):
	""" A data source which is simple wrapper around an in-memory array.
	"""

	###########################################################################
	@classmethod
	def default_chunk_size(cls):
		""" Returns the default chunk size for this source.
		"""
		return ChunkSource.USE_BATCH_SIZE

	###########################################################################
	def __init__(self, *sources, **kwargs):
		""" Create a new stacked data source.

			# Arguments

			*sources: list of Source instances. The data sources to stack.
		"""
		super().__init__(**kwargs)

		if not sources:
			raise ValueError('Cannot stack zero sources.')

		self.draw = None
		self.accumulator = None
		self.sources = []
		for source in sources:
			self.stack(source)

	###########################################################################
	def __len__(self):
		""" Returns the number of samples this source provides.
		"""
		return sum(len(source) for source in self.sources)

	###########################################################################
	def __getattr__(self, name):
		""" Pass-through for stacked sources.
		"""
		for source in self.sources:
			if hasattr(source, name):
				return getattr(source, name)
		raise AttributeError('No such attribute: {}'.format(name))

	###########################################################################
	def stack(self, source):
		""" Adds a new source to this data stack.
		"""
		if source.is_derived():
			raise ValueError('Cannot stack derived sources.')
		if not source:
			raise ValueError('Cannot stack sources with unknown lengths.')

		if self.sources:
			if len(source.shape()) != len(self.shape()):
				raise ValueError('Cannot stack sources with different shapes. '
					'We already had source(s) of shape {}, and now we were '
					'asked to stack a source of shape {}.'.format(
						self.shape(), source.shape()
					))
			for this_shape, that_shape in zip(self.shape(), source.shape()):
				if this_shape is None or that_shape is None:
					continue
				if this_shape != that_shape:
					raise ValueError('Incompatible shapes cannot be stacked. '
						'We already had source(s) of shape {}, and now we '
						'were asked to stack a source of shape {}.'.format(
							self.shape(), source.shape()
						))

		new_block = numpy.ones(len(source), dtype=numpy.int32) \
			* len(self.sources)
		if self.draw is None:
			self.draw = new_block
		else:
			self.draw = numpy.concatenate((self.draw, new_block))

		new_block = numpy.arange(len(source))
		if self.accumulator is None:
			self.accumulator = new_block
		else:
			self.accumulator = numpy.concatenate((self.accumulator, new_block))

		self.sources.append(source)

	###########################################################################
	def shape(self):
		""" Returns the shape of the individual data samples.
		"""
		return self.sources[0].shape()

	###########################################################################
	def can_shuffle(self):
		""" This source can be shuffled.
		"""
		return all(source.can_shuffle() for source in self.sources)

	###########################################################################
	def shuffle(self, indices):
		""" Applies a permutation to the data.
		"""
		if not self.can_shuffle():
			raise ValueError('This source is not shuffleable.')

		if len(indices) > len(self):
			raise ValueError('Shuffleable was asked to apply permutation, but '
				'the permutation is longer than the length of the data set.')

		indices = indices[:len(self)]

		self.draw = self.draw[indices]
		ctr = self.accumulator[indices]

		sub_shuffles = [[None]*len(source) for source in self.sources]
		counters = [0 for _ in range(len(self.sources))]
		for i, draw in enumerate(self.draw):
			self.accumulator[i] = counters[draw]
			sub_shuffles[draw][counters[draw]] = ctr[i]
			counters[draw] += 1

		for i, x in enumerate(sub_shuffles):
			self.sources[i].shuffle(x)

	###########################################################################
	def __iter__(self):
		""" Returns the next chunk of data.
		"""

		for source in self.sources:
			if isinstance(source, ChunkSource):
				if source.requested_chunk_size is ChunkSource.USE_BATCH_SIZE:
					source.set_chunk_size(self.chunk_size)

		queues = [None]*len(self.sources)
		iterators = [iter(source) for source in self.sources]

		assert self.draw is not None

		if isinstance(self.chunk_size, int):
			chunk_size = self.chunk_size
		else:
			chunk_size = ChunkSource.DEFAULT_CHUNK_SIZE

		start = 0
		num_entries = len(self)
		while start < num_entries:
			if self.chunk_size is ChunkSource.ALL_ITEMS:
				end = num_entries
			else:
				end = min(num_entries, start + chunk_size)

			num = end - start
			result = [None] * num
			for i in range(num):

				which = self.draw[start+i]

				if queues[which] is None:
					queues[which] = next(iterators[which])
				result[i] = queues[which][0]
				queues[which] = queues[which][1:]
				if len(queues[which]) == 0:	 # pylint: disable=len-as-condition
					queues[which] = None

			yield numpy.array(result)
			start = end

### EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF



"""
Copyright 2016 Deepgram

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import re
import logging
from ..utils import get_subclasses, CudaContext, CudaError

logger = logging.getLogger(__name__)

###############################################################################
class Backend:
	""" Base class for all tensor operation backends. Examples of backends are
		Keras, Theano, TensorFlow, etc.
	"""

	###########################################################################
	def __init__(self, variant=None, device=None, parallel=None):
		""" Create a new backend.

			Part of this call should be to ensure that all the necessary
			modules/libraries are available to use this backend. If this
			backend cannot be used, an exception should be raised. Otherwise,
			any appropriate configuration should take place here.

			# Arguments

			variant: str, list/tuple of str, or None (default: None). The
				variants to label this backend with. This allows for more
				streamlined and flexible modification of backend behavior at
				the layer level. These are simply tags that software components
				can query and modify their behavior in response, without
				requiring an entirely new backend.
		"""
		if not self.is_supported():
			logger.warning('Backend claims to not be supported. We will try '
				'to use it anyway.')

		if variant is not None:
			if isinstance(variant, str):
				variant = {variant}
		else:
			variant = set()

		self.variant = set(variant)

		if parallel is not None:
			if not isinstance(parallel, int):
				raise ValueError('If "parallel" is specified, it must be an '
					'integer. We received: {}'.format(parallel))
			if parallel <= 0:
				logger.warning('"parallel" should be > 0, but we received '
					'"%d". We are just ignoring "parallel".', parallel)
				parallel = None

		if device is None:
			logger.debug('No execution device indicated to backend. Checking '
				'available devices...')
			try:
				with CudaContext() as context:
					n_devices = len(context)
			except CudaError:
				logger.debug('Failed to initialize CUDA. Falling back to '
					'CPU.')
				device = 'cpu'
			else:
				if n_devices > 0:
					logger.debug('GPU capabilities detected.')
					device = 'gpu'
				else:
					logger.debug('No GPUs detected.')
					device = 'cpu'

		if device == 'cpu':
			self.devices = []
			if parallel:
				logger.warning('"parallel" execution was requested, but we '
					'not using the GPU. "parallel" will be ignored.')

		elif device.startswith('gpu'):
			if device == 'gpu':
				allowed = None
			else:
				allowed = set()
				regex = re.compile(
					r'gpu(?P<negate>!)?(?P<start>[0-9]+)(?:-(?P<end>[0-9]+))?')
				parts = device.split(',')
				for part in parts:
					match = regex.match(part.strip())
					if not match:
						raise ValueError('Failed to parse device: {}'.format(
							device))
					if match.group('end'):
						indices = set(range(
							int(match.group('start')),
							int(match.group('end'))+1
						))
					else:
						indices = {int(match.group('start'))}
					if match.group('negate') is None:
						allowed |= indices
					else:
						allowed -= indices

			with CudaContext() as context:
				n_devices = len(context)
				if allowed is None:
					allowed = set(range(n_devices))
				preferred = [
					(device.is_busy(), device.index()) \
						for device in context.rank_available()
				]
				preferred = [
					index for busy, index in preferred
					if index in allowed and not busy
				]

			if not preferred:
				logger.error('Failed to find any available GPUs.')
				raise ValueError('Failed to find any available GPUs.')

			if parallel:
				if parallel > len(preferred):
					logger.warning('%d GPU devices were requested through '
						'"parallel", but only %d devices seem '
						'available/unused. We are decreasing "parallel" to '
						'match.', parallel, len(preferred))
					self.devices = preferred
				else:
					self.devices = preferred[:parallel]
			else:
				self.devices = preferred

			logger.info('We are going to use %d GPU devices.',
				len(self.devices))

		else:
			raise ValueError('Invalid device specification: {}. If a '
				'device is explicitly specified, it must be "cpu", "gpu" '
				'or "gpuX" where "X" is an integer.'.format(device))

		logger.info('Creating backend: %s', self.get_name())
		logger.info('Backend variants: %s',
			'none' if not self.variant else ', '.join(
				str(x) for x in sorted(self.variant)
			)
		)

	###########################################################################
	@property
	def parallel(self):
		""" Returns the number of GPU devices we are using.
		"""
		if self.devices:
			return len(self.devices)
		return 0

	###########################################################################
	def has_variant(self, variant):
		""" Checks if a particular variant is enabled.
		"""
		return variant in self.variant

	###########################################################################
	@classmethod
	def is_supported(cls):
		""" Returns True if this backend can be used.

			# Return value

			If the backend appears to be useable, returns True. Otherwise,
			returns False.

			Note that if this returns False, then the backend should definitely
			not be able to be used. However, just because it returns True
			doesn't mean it will work (e.g., Theano with `force_device=true`,
			but without the requested device installed).
		"""
		raise NotImplementedError

	###########################################################################
	@staticmethod
	def from_specification(spec):
		""" Creates a new backend from the specification.

			# Arguments

			spec: str or dict. The backend specification. If processed from
				a standard job specification, it is everything under the
				`global.backend` section.

			# Return value

			Backend instance

			# Usage

			If no "global" section exists, or if no "backend" section exists in
			the "global" section, this will try to instantiate any supported
			backend installed on the system.

			Instantiate a specific backend with default parameters:
			```yaml
			global:
			  backend: BACKEND
			```
			or like this:
			```yaml
			global:
			  backend:
			    name: BACKEND
			```

			Instantiate a specific backend with additional keywords (the
			keywords must be supported by the Backend implementation):
			```yaml
			global:
			  backend:
			    name: BACKEND
				PARAM1: VALUE1
				PARAM2: VALUE2
				  ...
			```

			Instatiate any backend with additional keywords (the keywords must
			be supported by the Backend implementation):
			```yaml
			global:
			  backend:
				PARAM1: VALUE1
				PARAM2: VALUE2
				...
			```
		"""
		if spec is None:
			all_supported = list(Backend.get_all_backends(supported_only=True))
			if not all_supported:
				raise ValueError('No supported backends available.')
			target = all_supported[0]
			params = {}
		elif isinstance(spec, str):
			target = Backend.get_backend_by_name(spec)
			params = {}
		elif isinstance(spec, dict):
			if 'name' in spec:
				target = Backend.get_backend_by_name(spec.pop('name'))
			else:
				all_supported = list(Backend.get_all_backends(
					supported_only=True))
				if not all_supported:
					raise ValueError('No supported backends available.')
				target = all_supported[0]
			params = spec
		else:
			raise ValueError(
				'Unexpected backend specification: {}'.format(spec))

		logger.debug('Using backend: %s', target.get_name())
		result = target(**params)
		return result

	###########################################################################
	@staticmethod
	def get_backend_by_name(name):
		""" Factory method for creating new backends.

			# Arguments

			name: str. The name of the backend to instantiate. This is given by
				the backend's `get_name()` method.
			kwargs: dict. The arguments to pass to the backend's constructor.

			# Return value

			If the named backend is found, this returns a Backend instance;
			otherwise, a ValueError is raised.
		"""
		name = name.lower()
		for cls in Backend.get_all_backends(supported_only=False):
			if cls.get_name() == name:
				return cls
		raise ValueError('No such backend: {}'.format(name))

	###########################################################################
	@staticmethod
	def get_all_backends(supported_only=False):
		""" Factory method for creating backend which is supported on the host
			platform.

			# Arguments

			kwargs: dict. The arguments to pass to the backend's constructor.

			# Return value

			If the named backend is found, this returns a Backend instance;
			otherwise, a ValueError is raised.
		"""
		for cls in get_subclasses(Backend):
			if supported_only:
				if cls.is_supported():
					yield cls
			else:
				yield cls

	###########################################################################
	@staticmethod
	def get_any_supported_backend():
		""" Finds a backend class that claims to be supported on this system.
		"""
		for cls in Backend.get_all_backends(supported_only=True):
			return cls
		raise ValueError('No supported backends found.')

	###########################################################################
	@classmethod
	def get_name(cls):
		""" Return a lower-case string naming this backend.
		"""
		return cls.__name__.lower()

	###########################################################################
	def save(self, model, filename):
		""" Saves the model weights to the given filename.

			# Arguments

			model: Model instance. The model whose weights should be saved.
			filename: str. The filename to write the weights to.

			# Notes

			The file format is backend-specific. There is no guarantee of
			compatability between different backends.

			# Return value

			None
		"""
		raise NotImplementedError

	###########################################################################
	def restore(self, model, filename):
		""" Load the model weights from the given filename.

			# Arguments

			model: Model instance. The model whose weights should be restored.
			filename: str. The filename to read the weights from.

			# Notes

			The file format is backend-specific. There is no guarantee of
			compatability between different backends.

			# Return value

			None
		"""
		raise NotImplementedError

	###########################################################################
	def create_data(self, model):
		""" Requests a new set of model-specific data to be used during model
			assembly.
		"""
		return None

	###########################################################################
	def connect(self, inputs, target, data):
		""" Applies a tensor operation to a set of input tensors.

			# Arguments

			inputs: list. A list of input tensors.
			target: object. The tensor operation to apply.
			data: object. A backend-specific object, created by
				`create_data()`, which can be used by the backend to maintain
				state while constructing the model.

			# Notes

			The exact underlying types of the `inputs` and `target` tensors
			depend on the backend implementation.

			# Return value

			A new tensor object (backend-specific) resulting from the applied
			operation.
		"""
		raise NotImplementedError

	###########################################################################
	def compile(self, model, loss=None, optimizer=None, blocking=True,
		assemble_only=False):
		""" Creates an implementation-specific representation of the
			instantiated model.

			# Arguments

			model: Model instance. The built model to create a representation
				for.
			loss: Loss instance, list, dict, or None. The loss function
				(objective function) to use for training. If None, no loss
				functions are used. If a list, it should be a list of Loss
				instances, one for each model output; if the list is length 1,
				then the same loss function is used for each model output. If
				this is a dict, it should map from model output names to Loss
				instances to apply to that output.
			optimizer: Optimizer instance or None. The loss function to use for
				training. If None, no optimizer is used.
			blocking: bool (default: True). If True, requests that the
				implementation not return until all code compiling is complete.
				If False, the implementation may spawn background threads or
				processes, and these may still be running when
				train/test/evaluate are called (leading to blocking behavior at
				that point).
			assemble_only: bool (default: False). If True, does not attempt to
				create any compiled functions/models at all. Simply assembles
				the model and prepares the data sources. If this is True, then
				blocking is ignored.

			# Return value

			An implementation-specific object ("model") that can be used for
			training or evaluation. However, if `assemble_only` is True, then
			this returns a limited object which cannot be actually used for
			anything.

			# Notes

			Implementations may need to modify the Model. If this is the case,
			ExtensionState should be used to protect the model from extensions
			that are left stuck on after compiling.
		"""
		raise NotImplementedError

	###########################################################################
	def train(self, model, data):
		""" Fits the given model on a batch of data.

			# Arguments

			model: Model instance. The model to train with.
			data: dict or list. The data to train a batch on.
			compiled: object. The data package returned by `compile`. Its type
				and content are backend-specific.

			# Return value

			A dictionary of loss values, whose keys are the names of the
			respective outputs. If per-output loss is not available, then a
			single, global loss value can be returned instead.
		"""
		raise NotImplementedError

	###########################################################################
	def test(self, model, data):
		""" Calculates the model loss on a batch of data.

			# Arguments

			model: Model instance. The model to evaluate with.
			data: dict or list. The batch of data to test on.
			compiled: object. The data package returned by `compile`. Its type
				and content are backend-specific.

			# Return value

			A dictionary of loss values, whose keys are the names of the
			respective outputs. If per-output loss is not available, then a
			single, global loss value can be returned instead.
		"""
		raise NotImplementedError

	###########################################################################
	def evaluate(self, model, data):
		""" Evaluates the model on a batch ofdata.
		"""
		raise NotImplementedError

### EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF



"""
Copyright 2017 Deepgram

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import os
import logging

import urllib.request

from . import TrainingHook, PlotHook
from ...utils import prepare_json

from tempfile import NamedTemporaryFile

logger = logging.getLogger(__name__)

###############################################################################
class KurhubHook(TrainingHook):
    """ Hook for posting to Kurhub.
    """

    ###########################################################################
    @classmethod
    def get_name(cls):
        """ Returns the name of the hook.
        """
        return 'kurhub'

    ###########################################################################
    def __init__(self, uuid=None, endpoint="http://dev.kurhub.com/kur_updates", *args, **kwargs):
        """ Creates a new kurhub hook.
        """
        super().__init__(*args, **kwargs)

        self.uuid = uuid
        self.endpoint = endpoint

        with NamedTemporaryFile() as tmp_file:
            tmp_file.close()
            self.plot_name = tmp_file.name 

        self.plot_hook = PlotHook(self.plot_name)


        if uuid is None:
            raise ValueError('kurhub hook requires a "uuid" to be defined.')

    ###########################################################################
    def send_message(self, text, info=None):
        """ Sends a message to kurhub.
        """
        data = {
            'text': text,
            'uuid': self.uuid
        }

        url = self.endpoint
        data, header = prepare_json(data)
        self._submit(url, data, header)

    ###########################################################################     
    def send_plot_message(self, text, plot_string, info=None):
        """ Sends a plot message to kurhub.
        """
        data = {
            'text': text,
            'plot': plot_string,
            'uuid': self.uuid
        }

        url = self.endpoint
        data, header = prepare_json(data)
        self._submit(url, data, header)

    ###########################################################################
    def _submit(self, url, data, header):
        """ Submits a POST request to kurhub.
        """
        request = urllib.request.Request(
            url,
            data=data,
            headers=header,
            method='POST'
        )
        director = urllib.request.build_opener()

        try:
            response = director.open(request)
        except:                                 # pylint: disable=bare-except
            logger.exception('Failed to connect to kurhub. Make sure the URL '
                'and channel are correct. If the channel was newly created, '
                'it might take a little time for kurhub to catch up.')
        else:
            if response.code != 200:
                logger.error('Failed to post kurhub notification. Make sure '
                'the URL and channel are correct. If the channel was newly '
                'created, it might take a little time for kurhub to catch up.')

    ###########################################################################
    def notify(self, status, log=None, info=None):
        """ Sends the kurhub message.
        """
        
        self.plot_hook.notify(status, log, info)
        # check if plot
        plot_name = '{}.png'.format(self.plot_name)
        if os.path.isfile(plot_name):
            if os.stat(plot_name).st_size > 0:
                # upload
                with open(plot_name, 'rb') as plotfile:
                    import base64
                    encoded_string = base64.b64encode(plotfile.read()).decode('utf-8')

                ## send as post request base64
                self.send_plot_message('plot created', encoded_string)
                # delete after upload
                os.remove(plot_name)

        info = info or {}

        if status is TrainingHook.EPOCH_END:
            epoch = info.pop('epoch', None)
            total_epochs = info.pop('total_epochs', None)
            text = 'Finished epoch {} of {}.'.format(epoch, total_epochs)
        elif status is TrainingHook.TRAINING_END:
            text = 'Training has ended.'
        elif status is TrainingHook.TRAINING_START:
            text = 'Started training.'
        else:
            text = None

        if text:
            self.send_message(text, info)

### EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF
